import logging as _logging
import os as _os
from pathlib import Path

import click

from flytekit.clis.sdk_in_container.constants import CTX_PACKAGES
from flytekit.clis.sdk_in_container.fast_register import fast_register
from flytekit.clis.sdk_in_container.launch_plan import launch_plans
from flytekit.clis.sdk_in_container.register import register
from flytekit.clis.sdk_in_container.serialize import serialize
from flytekit.configuration import internal as _internal_config
from flytekit.configuration import platform as _platform_config
from flytekit.configuration import sdk as _sdk_config
from flytekit.configuration import set_flyte_config_file
from flytekit.configuration.internal import CONFIGURATION_PATH
from flytekit.configuration.platform import URL as _URL
from flytekit.configuration.sdk import WORKFLOW_PACKAGES as _WORKFLOW_PACKAGES


@click.group("pyflyte", invoke_without_command=True)
@click.option(
    "-c",
    "--config",
    required=False,
    type=str,
    help="Path to config file for use within container",
)
@click.option(
    "-k",
    "--pkgs",
    required=False,
    multiple=True,
    help="Dot separated python packages to operate on.  Multiple may be specified  Please note that this "
    "option will override the option specified in the configuration file, or environment variable",
)
@click.option(
    "-i",
    "--insecure",
    required=False,
    type=bool,
    help="Do not use SSL to connect to Admin",
)
@click.pass_context
def main(ctx, config=None, pkgs=None, insecure=None):
    """
    Entrypoint for all the user commands.
    """
    update_configuration_file(config)

    # Update the logger if it's set
    log_level = _internal_config.LOGGING_LEVEL.get() or _sdk_config.LOGGING_LEVEL.get()
    if log_level is not None:
        _logging.getLogger().setLevel(log_level)

    ctx.obj = dict()

    # Determine SSL.  Note that the insecure option in this command is not a flag because we want to look
    # up configuration settings if it's missing.  If the command line option says insecure but the config object
    # says no, let's override the config object by overriding the environment variable.
    if insecure and not _platform_config.INSECURE.get():
        _platform_config.INSECURE.get()
        _os.environ[_platform_config.INSECURE.env_var] = "True"

    # Handle package management - get from config if not specified on the command line
    pkgs = pkgs or []
    if len(pkgs) == 0:
        pkgs = _WORKFLOW_PACKAGES.get()
    ctx.obj[CTX_PACKAGES] = pkgs


def update_configuration_file(config_file_path):
    """
    Changes the configuration singleton object to read from another file if specified, which should be
    at the base of the repository.

    :param Text config_file_path:
    """
    configuration_file = Path(config_file_path or CONFIGURATION_PATH.get())
    if configuration_file.is_file():
        click.secho(
            "Using configuration file at {}".format(configuration_file.absolute().as_posix()),
            fg="green",
        )
        set_flyte_config_file(configuration_file.as_posix())
    else:
        click.secho(
            "Configuration file '{}' could not be loaded. Using values from environment.".format(
                CONFIGURATION_PATH.get()
            ),
            color="yellow",
        )
        set_flyte_config_file(None)
    click.secho("Flyte Admin URL {}".format(_URL.get()), fg="green")


main.add_command(register)
main.add_command(fast_register)
main.add_command(serialize)
main.add_command(launch_plans)

if __name__ == "__main__":
    main()



import datetime as _datetime

import six as _six

from flytekit.common import constants as _common_constants
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.tasks import generic_spark_task as _sdk_generic_spark_task
from flytekit.common.tasks import hive_task as _sdk_hive_tasks
from flytekit.common.tasks import pytorch_task as _sdk_pytorch_tasks
from flytekit.common.tasks import sdk_dynamic as _sdk_dynamic
from flytekit.common.tasks import sdk_runnable as _sdk_runnable_tasks
from flytekit.common.tasks import sidecar_task as _sdk_sidecar_tasks
from flytekit.common.tasks import spark_task as _sdk_spark_tasks
from flytekit.common.tasks import task as _task
from flytekit.common.tasks import tensorflow_task as _sdk_tensorflow_tasks
from flytekit.common.types import helpers as _type_helpers
from flytekit.contrib.notebook import tasks as _nb_tasks
from flytekit.models import interface as _interface_model
from flytekit.sdk.spark_types import SparkType as _spark_type


def inputs(_task_template=None, **kwargs):
    """
    Decorator that provides input definitions to a decorated task definition.

    .. note::

        Certain tasks have special input behavior.  See comments on each task decorator for more information.

    .. code-block:: python

        @inputs(in1=Types.Integer, in2=[Types.String], in3=[[[Types.Integer]]])
        @outputs(out1=Types.Integer, out2=Types.String)
        @python_task
        def my_task(wf_params, in1, in2, out1, out2):
            pass

    :param flytekit.common.tasks.sdk_runnable.SdkRunnableTask _task_template: Do not declare directly.  This is the
        decorated task template.
    :param dict[Text,flytekit.common.types.base_sdk_types.FlyteSdkType] kwargs: Arbitrary keyword arguments for input
        name and type.
    :rtype: flytekit.common.tasks.sdk_runnable.SdkRunnableTask
    """

    def apply_inputs_wrapper(task):
        if not isinstance(task, _task.SdkTask):
            additional_msg = (
                "Inputs can only be applied to a task. Did you forget the task decorator on method '{}.{}'?".format(
                    task.__module__,
                    task.__name__ if hasattr(task, "__name__") else "<unknown>",
                )
            )
            raise _user_exceptions.FlyteTypeException(
                expected_type=_sdk_runnable_tasks.SdkRunnableTask,
                received_type=type(task),
                received_value=task,
                additional_msg=additional_msg,
            )
        for k, v in _six.iteritems(kwargs):
            kwargs[k] = _interface_model.Variable(
                _type_helpers.python_std_to_sdk_type(v).to_flyte_literal_type(), ""
            )  # TODO: Support descriptions

        task.add_inputs(kwargs)
        return task

    if _task_template is not None:
        return apply_inputs_wrapper(_task_template)
    else:
        return apply_inputs_wrapper


def outputs(_task_template=None, **kwargs):
    """
    Decorator that provides output definitions to a decorated task definition.

    .. note::

        Certain tasks have special output behavior.  See comments on each task decorator for more information.

    .. code-block:: python

        @outputs(out1=Types.Integer, out2=Types.String)
        @python_task
        def my_task(wf_params, out1, out2):
            out1.set(123)
            out2.set('hello world!')

    :param flytekit.common.tasks.sdk_runnable.SdkRunnableTask _task_template: Do not declare directly.  This is the
        decorated task template.
    :param dict[Text,flytekit.common.types.base_sdk_types.FlyteSdkType] kwargs: Arbitrary keyword arguments for input
        name and type.
    :rtype: flytekit.common.tasks.sdk_runnable.SdkRunnableTask
    """

    def apply_outputs_wrapper(task):
        if not isinstance(task, _sdk_runnable_tasks.SdkRunnableTask) and not isinstance(
            task, _nb_tasks.SdkNotebookTask
        ):
            additional_msg = (
                "Outputs can only be applied to a task. Did you forget the task decorator on method '{}.{}'?".format(
                    task.__module__,
                    task.__name__ if hasattr(task, "__name__") else "<unknown>",
                )
            )
            raise _user_exceptions.FlyteTypeException(
                expected_type=_sdk_runnable_tasks.SdkRunnableTask,
                received_type=type(task),
                received_value=task,
                additional_msg=additional_msg,
            )
        for k, v in _six.iteritems(kwargs):
            kwargs[k] = _interface_model.Variable(
                _type_helpers.python_std_to_sdk_type(v).to_flyte_literal_type(), ""
            )  # TODO: Support descriptions

        task.add_outputs(kwargs)
        return task

    if _task_template is not None:
        return apply_outputs_wrapper(_task_template)
    else:
        return apply_outputs_wrapper


def python_task(
    _task_function=None,
    cache_version="",
    retries=0,
    interruptible=None,
    deprecated="",
    storage_request=None,
    cpu_request=None,
    gpu_request=None,
    memory_request=None,
    storage_limit=None,
    cpu_limit=None,
    gpu_limit=None,
    memory_limit=None,
    cache=False,
    timeout=None,
    environment=None,
    cls=None,
):
    """
    Decorator to create a Python Task definition.  This task will run as a single unit of work on the platform.

    .. code-block:: python

        @inputs(int_list=[Types.Integer])
        @outputs(sum_of_list=Types.Integer
        @python_task
        def my_task(wf_params, int_list, sum_of_list):
            sum_of_list.set(sum(int_list))

    :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed for wrapped task functions.

    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.

        .. note::

            This argument is required to be a non-empty string if `cache` is True.

    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.sdk.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.

        .. note::

            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.

    :param bool interruptible: [optional] boolean describing if the task is interruptible.

    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.

    :param Text storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for the task to run.  Default is set by platform-level configuration.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_request: [optional] Kubernetes resource string for lower-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs.
        Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for the task to execute.  Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for the task to run.  This amount is not guaranteed!  If not specified, it is set equal to storage_request.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.

    :param Text gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs. This amount is not
        guaranteed!  If not specified, it is set equal to gpu_request.

    :param Text memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for the task to execute.  This amount is not guaranteed!  If not specified, it is set equal to
        memory_request.

    :param bool cache: [optional] boolean describing if the outputs of this task should be cached and
        re-usable.

    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.

    :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.

    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided must be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask.  A user can use this to
        inject bespoke logic into the base Flyte programming model.

    :rtype: flytekit.common.tasks.sdk_runnable.SdkRunnableTask
    """

    def wrapper(fn):
        return (cls or _sdk_runnable_tasks.SdkRunnableTask)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.PYTHON_TASK,
            discovery_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            storage_request=storage_request,
            cpu_request=cpu_request,
            gpu_request=gpu_request,
            memory_request=memory_request,
            storage_limit=storage_limit,
            cpu_limit=cpu_limit,
            gpu_limit=gpu_limit,
            memory_limit=memory_limit,
            discoverable=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            environment=environment,
            custom={},
        )

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper


def dynamic_task(
    _task_function=None,
    cache_version="",
    retries=0,
    interruptible=None,
    deprecated="",
    storage_request=None,
    cpu_request=None,
    gpu_request=None,
    memory_request=None,
    storage_limit=None,
    cpu_limit=None,
    gpu_limit=None,
    memory_limit=None,
    cache=False,
    timeout=None,
    allowed_failure_ratio=None,
    max_concurrency=None,
    environment=None,
    cls=None,
):
    """
    Decorator to create a custom dynamic task definition.  Dynamic tasks should be used to split up work into
    an arbitrary number of parallel sub-tasks, or workflows.

    .. code-block:: python

        @outputs(out=Types.Integer)
        @python_task
        def my_sub_task(wf_params, out):
            out.set(randint())

        @outputs(out=[Types.Integer])
        @dynamic_task
        def my_task(wf_params, out):
            out_list = []
            for i in xrange(100):
                out_list.append(my_sub_task().outputs.out)
            out.set(out_list)

    .. note::

        All outputs of a batch task must be a list.  This is because the individual outputs of sub-tasks should be
        appended into a list.  There cannot be aggregation of outputs done in this task.  To accomplish aggregation,
        it is recommended that a python_task take the outputs of this task as input and do the necessary work.
        If a sub-task does not contribute an output, it must be yielded from the task with the `yield` keyword or
        returned from the task in a list.  If this isn't done, the sub-task will not be executed.

    :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed.
    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.

        .. note::

            This argument is required to be a non-empty string if `cache` is True.

    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.sdk.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.

        .. note::

            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.

    :param bool interruptible: [optional] boolean describing if the task is interruptible.

    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.
    :param Text storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for the task to run.  Default is set by platform-level configuration.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_request: [optional] Kubernetes resource string for lower-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs.
        Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for the task to execute.  Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for the task to run.  This amount is not guaranteed!  If not specified, it is set equal to storage_request.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.
    :param Text gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs. This amount is not
        guaranteed!  If not specified, it is set equal to gpu_request.
    :param Text memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for the task to execute.  This amount is not guaranteed!  If not specified, it is set equal to
        memory_request.
    :param bool cache: [optional] boolean describing if the outputs of this task should be cached and
        re-usable.
    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.
    :param float allowed_failure_ratio: [optional] float value describing the ratio of sub-tasks that may fail before
        the master batch task considers itself a failure.  By default, the value is 0 so if any task fails, the master
        batch task will be marked a failure.  If specified, the value must be between 0 and 1 inclusive.  In the event a
        non-zero value is specified, downstream tasks must be able to accept None values as outputs from individual
        sub-tasks because the output values will be set to None for any sub-task that fails.
    :param int max_concurrency: [optional] integer value describing the maximum number of tasks to run concurrently.
        This is a stand-in pending better concurrency controls for special use-cases.  The existence of this parameter
        is not guaranteed between versions and therefore it is NOT recommended that it be used.
    :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.
    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided must be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask.  Generally, it should be a
        subclass of flytekit.common.tasks.sdk_dynamic.SdkDynamicTask.  A user can use this parameter to inject bespoke
        logic into the base Flyte programming model.
    :rtype: flytekit.common.tasks.sdk_runnable.SdkDynamicTask
    """

    def wrapper(fn):
        return (cls or _sdk_dynamic.SdkDynamicTask)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.DYNAMIC_TASK,
            discovery_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            storage_request=storage_request,
            cpu_request=cpu_request,
            gpu_request=gpu_request,
            memory_request=memory_request,
            storage_limit=storage_limit,
            cpu_limit=cpu_limit,
            gpu_limit=gpu_limit,
            memory_limit=memory_limit,
            discoverable=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            allowed_failure_ratio=allowed_failure_ratio,
            max_concurrency=max_concurrency,
            environment=environment or {},
            custom={},
        )

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper


def spark_task(
    _task_function=None,
    cache_version="",
    retries=0,
    interruptible=None,
    deprecated="",
    cache=False,
    timeout=None,
    spark_conf=None,
    hadoop_conf=None,
    environment=None,
    cls=None,
):
    """
    Decorator to create a spark task.  This task will connect to a Spark cluster, configure the environment,
    and then execute the code within the _task_function as the Spark driver program.

    .. code-block:: python

        @inputs(a=Types.Integer)
        @spark_task(
            spark_conf={
                    'spark.executor.cores': '7',
                    'spark.executor.instances': '31',
                    'spark.executor.memory': '32G'
                }
            )
        def sparky(wf_params, spark_context, a):
            pass

    :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed for wrapped task functions.
    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.

        .. note::

            This argument is required to be a non-empty string if `cache` is True.

    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.sdk.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.

        .. note::

            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.

    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.
    :param bool cache: [optional] boolean describing if the outputs of this task should be cached and
        re-usable.
    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.
    :param dict[Text,Text] spark_conf: A definition of key-value pairs for spark config for the job.
    :param dict[Text,Text] hadoop_conf: A definition of key-value pairs for hadoop config for the job.
    :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.
    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided must be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask.  Generally, it should be a
        subclass of flytekit.common.tasks.spark_task.SdkSparkTask.  A user can use this parameter to inject bespoke
        logic into the base Flyte programming model.
    :rtype: flytekit.common.tasks.sdk_runnable.SdkRunnableTask
    """

    def wrapper(fn):
        return (cls or _sdk_spark_tasks.SdkSparkTask)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.SPARK_TASK,
            discovery_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            spark_type=_spark_type.PYTHON,
            deprecated=deprecated,
            discoverable=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            spark_conf=spark_conf or {},
            hadoop_conf=hadoop_conf or {},
            environment=environment or {},
        )

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper


def generic_spark_task(
    spark_type,
    main_class,
    main_application_file,
    cache_version="",
    retries=0,
    interruptible=None,
    inputs=None,
    deprecated="",
    cache=False,
    timeout=None,
    spark_conf=None,
    hadoop_conf=None,
    environment=None,
):
    """
    Create a generic spark task. This task will connect to a Spark cluster, configure the environment,
    and then execute the mainClass code as the Spark driver program.

    """

    return _sdk_generic_spark_task.SdkGenericSparkTask(
        task_type=_common_constants.SdkTaskType.SPARK_TASK,
        discovery_version=cache_version,
        retries=retries,
        interruptible=interruptible,
        deprecated=deprecated,
        discoverable=cache,
        timeout=timeout or _datetime.timedelta(seconds=0),
        spark_type=spark_type,
        task_inputs=inputs,
        main_class=main_class or "",
        main_application_file=main_application_file or "",
        spark_conf=spark_conf or {},
        hadoop_conf=hadoop_conf or {},
        environment=environment or {},
    )


def qubole_spark_task(*args, **kwargs):
    """
    :rtype: flytekit.common.tasks.sdk_runnable.SdkRunnableTask
    """
    raise NotImplementedError("Qubole Spark Task is currently not supported in Flyte.")


def hive_task(
    _task_function=None,
    cache_version="",
    retries=0,
    interruptible=None,
    deprecated="",
    storage_request=None,
    cpu_request=None,
    gpu_request=None,
    memory_request=None,
    storage_limit=None,
    cpu_limit=None,
    gpu_limit=None,
    memory_limit=None,
    cache=False,
    timeout=None,
    environment=None,
    cls=None,
):
    """
    Decorator to create a hive task. This task should output a list of hive queries which are run on a hive cluster.

    This is a 2 step task:

        1. Generator step runs in the user container and outputs a list of queries.
        2. The list of queries produced in step1 is then submitted to Hive Cluster. The queries are monitored by Flyte
           Backend for completion.
    Container properties(cpu, gpu, memory, etc) set on this task are only used in step1 above.

    .. code-block:: python

        @inputs(a=Types.Integer)
        @hive_task(
            cache_version='1',
        )
        def test_hive(wf_params, a):
            return [
                "SELECT * FROM users_table WHERE user_id=4",
                "INSERT INTO users_table VALUES ("user", 5, 4)"
            ]

     :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed for wrapped task functions.

    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.

        .. note::

            This argument is required to be a non-empty string if `cache` is True.

    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.common.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.

        .. note::

            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.

    :param bool interruptible: [optional] boolean describing if task is interruptible.
    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.
    :param Text storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for the task to run.  Default is set by platform-level configuration.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_request: [optional] Kubernetes resource string for lower-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs.
        Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for the task to execute.  Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for the task to run.  This amount is not guaranteed!  If not specified, it is set equal to storage_request.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.
    :param Text gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs. This amount is not
        guaranteed!  If not specified, it is set equal to gpu_request.
    :param Text memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for the task to execute.  This amount is not guaranteed!  If not specified, it is set equal to
        memory_request.
    :param bool cache: [optional] boolean describing if the outputs of this task should be cached and
        re-usable.
    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.
    :param dict[Text,Text] environment: Environment variables to set for the execution of the query-generating
        container.
    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided should be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask. Generally, it should be
        a subclass of flytekit.common.tasks.hive_task.SdkHiveTask. A user can use this to inject bespoke logic into
        the base Flyte programming model.

    :rtype: flytekit.common.tasks.sdk_runnable.SdkHiveTask
    """

    def wrapper(fn):

        return (cls or _sdk_hive_tasks.SdkHiveTask)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.BATCH_HIVE_TASK,
            discovery_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            storage_request=storage_request,
            cpu_request=cpu_request,
            gpu_request=gpu_request,
            memory_request=memory_request,
            storage_limit=storage_limit,
            cpu_limit=cpu_limit,
            gpu_limit=gpu_limit,
            memory_limit=memory_limit,
            discoverable=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            cluster_label="",
            tags=[],
            environment=environment or {},
        )

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper


def qubole_hive_task(
    _task_function=None,
    cache_version="",
    retries=0,
    interruptible=None,
    deprecated="",
    storage_request=None,
    cpu_request=None,
    gpu_request=None,
    memory_request=None,
    storage_limit=None,
    cpu_limit=None,
    gpu_limit=None,
    memory_limit=None,
    cache=False,
    timeout=None,
    cluster_label=None,
    tags=None,
    environment=None,
    cls=None,
):
    """
    Decorator to create a qubole hive task. This is hive task runs on a qubole cluster, and therefore allows users to
    pass cluster labels and qubole query tags. Similar to hive task, this task should output a list of hive queries
    that are run on a hive cluster.

    Similar to a hive task, this is also a 2 step task where step2 is run on a qubole hive cluster. Therefore, users can
    specify qubole cluster_label and query tags on this task.

    .. code-block:: python

        @inputs(a=Types.Integer)
        @qubole_hive_task(
            cache_version='1',
            cluster_label='cluster_label',
            tags=['tag1'],
            )
        def test_hive(wf_params, a):
            return [
                "SELECT * FROM users_table WHERE user_id=4",
                "INSERT INTO users_table VALUES ("user", 5, 4)"
            ]

     :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed for wrapped task functions.

    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.

        .. note::

            This argument is required to be a non-empty string if `cache` is True.

    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.common.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.

        .. note::

            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.

    :param bool interruptible: [optional] boolean describing if task is interruptible.
    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.
    :param Text storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for the task to run.  Default is set by platform-level configuration.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_request: [optional] Kubernetes resource string for lower-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs.
        Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for the task to execute.  Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for the task to run.  This amount is not guaranteed!  If not specified, it is set equal to storage_request.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.
    :param Text gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs. This amount is not
        guaranteed!  If not specified, it is set equal to gpu_request.
    :param Text memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for the task to execute.  This amount is not guaranteed!  If not specified, it is set equal to
        memory_request.
    :param bool cache: [optional] boolean describing if the outputs of this task should be cached and
        re-usable.
    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.
    :param cluster_label: The qubole cluster label where the query is to be executed
    :param list[Text] tags: User defined tags(key-value pairs) defined by the user for the queries. These tags are
        passed to Qubole.
    :param dict[Text,Text] environment: Environment variables to set for the execution of the query-generating
        container.
    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided should be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask. Generally, it should be
        a subclass of flytekit.common.tasks.hive_task.SdkHiveTask. A user can use this to inject bespoke logic into
        the base Flyte programming model.

    :rtype: flytekit.common.tasks.sdk_runnable.SdkHiveTask
    """

    def wrapper(fn):

        return (cls or _sdk_hive_tasks.SdkHiveTask)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.BATCH_HIVE_TASK,
            discovery_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            storage_request=storage_request,
            cpu_request=cpu_request,
            gpu_request=gpu_request,
            memory_request=memory_request,
            storage_limit=storage_limit,
            cpu_limit=cpu_limit,
            gpu_limit=gpu_limit,
            memory_limit=memory_limit,
            discoverable=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            cluster_label=cluster_label or "",
            tags=tags or [],
            environment=environment or {},
        )

    # This is syntactic-sugar, so that when calling this decorator without args, you can either
    # do it with () or without any ()
    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper


def sidecar_task(
    _task_function=None,
    cache_version="",
    retries=0,
    interruptible=None,
    deprecated="",
    storage_request=None,
    cpu_request=None,
    gpu_request=None,
    memory_request=None,
    storage_limit=None,
    cpu_limit=None,
    gpu_limit=None,
    memory_limit=None,
    cache=False,
    timeout=None,
    environment=None,
    pod_spec=None,
    primary_container_name=None,
    cls=None,
):
    """
    Decorator to create a Sidecar Task definition.  This task will execute the primary task alongside the specified
    kubernetes PodSpec. Custom primary task container attributes can be defined in the PodSpec by defining a container
    whose name matches the primary_container_name. These container attributes will be applied to the container brought
    up to execute the primary task definition.

    .. code-block:: python

        def generate_pod_spec_for_task():
            pod_spec = generated_pb2.PodSpec()
            secondary_container = generated_pb2.Container(
                name="secondary",
                image="alpine",
            )
            secondary_container.command.extend(["/bin/sh"])
            secondary_container.args.extend(["-c", "echo hi sidecar world > /data/message.txt"])
            shared_volume_mount = generated_pb2.VolumeMount(
                name="shared-data",
                mountPath="/data",
            )
            secondary_container.volumeMounts.extend([shared_volume_mount])

            primary_container = generated_pb2.Container(name="primary")
            primary_container.volumeMounts.extend([shared_volume_mount])

            pod_spec.volumes.extend([generated_pb2.Volume(
                name="shared-data",
                volumeSource=generated_pb2.VolumeSource(
                    emptyDir=generated_pb2.EmptyDirVolumeSource(
                        medium="Memory",
                    )
                )
            )])
            pod_spec.containers.extend([primary_container, secondary_container])
            return pod_spec

        @sidecar_task(
            pod_spec=generate_pod_spec_for_task(),
            primary_container_name="primary",
        )
        def a_sidecar_task(wfparams):
            while not os.path.isfile('/data/message.txt'):
                time.sleep(5)


    :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed for wrapped task functions.

    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.
        .. note::
            This argument is required to be a non-empty string if `cache` is True.

    :param int retries: [optional] integer determining number of times task can be retried on
        :py:ex:`flytekit.sdk.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.

        .. note::

            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.

    :param bool interruptible: Specify whether task is interruptible

    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.

    :param Text storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for the task to run.  Default is set by platform-level configuration.

        TODO: !!!
        .. note::

            This is currently not supported by the platform.

    :param Text cpu_request: [optional] Kubernetes resource string for lower-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs.
        Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for the task to execute.  Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for the task to run.  This amount is not guaranteed!  If not specified, it is set equal to storage_request.

        TODO: !!!
        .. note::

            This is currently not supported by the platform.

    :param Text cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.

    :param Text gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs. This amount is not
        guaranteed!  If not specified, it is set equal to gpu_request.

    :param Text memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for the task to execute.  This amount is not guaranteed!  If not specified, it is set equal to
        memory_request.

    :param bool cache: [optional] boolean describing if the outputs of this task should be discoverable and
        re-usable.

    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.

    :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.

    :param k8s.io.api.core.v1.generated_pb2.PodSpec pod_spec: [optional] PodSpec to bring up alongside task execution.

    :param Text primary_container_name: primary container to monitor for the duration of the task.

    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided must be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask.  A user can use this to
        inject bespoke logic into the base Flyte programming model.

    :rtype: flytekit.common.tasks.sdk_runnable.SdkRunnableTask

    """

    def wrapper(fn):

        return (cls or _sdk_sidecar_tasks.SdkSidecarTask)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.SIDECAR_TASK,
            discovery_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            storage_request=storage_request,
            cpu_request=cpu_request,
            gpu_request=gpu_request,
            memory_request=memory_request,
            storage_limit=storage_limit,
            cpu_limit=cpu_limit,
            gpu_limit=gpu_limit,
            memory_limit=memory_limit,
            discoverable=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            environment=environment,
            pod_spec=pod_spec,
            primary_container_name=primary_container_name,
        )

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper


def dynamic_sidecar_task(
    _task_function=None,
    cache_version="",
    retries=0,
    interruptible=None,
    deprecated="",
    storage_request=None,
    cpu_request=None,
    gpu_request=None,
    memory_request=None,
    storage_limit=None,
    cpu_limit=None,
    gpu_limit=None,
    memory_limit=None,
    cache=False,
    timeout=None,
    allowed_failure_ratio=None,
    max_concurrency=None,
    environment=None,
    pod_spec=None,
    primary_container_name=None,
    cls=None,
):
    """
    Decorator to create a custom dynamic sidecar task definition. Dynamic
    tasks should be used to split up work into an arbitrary number of parallel
    sub-tasks, or workflows. This task will execute the primary task alongside
    the specified kubernetes PodSpec. Custom primary task container attributes
    can be defined in the PodSpec by defining a container whose name matches
    the primary_container_name. These container attributes will be applied to
    the container brought up to execute the primary task definition.
    .. code-block:: python
        def generate_pod_spec_for_task():
            pod_spec = generated_pb2.PodSpec()
            secondary_container = generated_pb2.Container(
                name="secondary",
                image="alpine",
            )
            secondary_container.command.extend(["/bin/sh"])
            secondary_container.args.extend(["-c", "echo hi sidecar world > /data/message.txt"])
            shared_volume_mount = generated_pb2.VolumeMount(
                name="shared-data",
                mountPath="/data",
            )
            secondary_container.volumeMounts.extend([shared_volume_mount])
            primary_container = generated_pb2.Container(name="primary")
            primary_container.volumeMounts.extend([shared_volume_mount])
            pod_spec.volumes.extend([generated_pb2.Volume(
                name="shared-data",
                volumeSource=generated_pb2.VolumeSource(
                    emptyDir=generated_pb2.EmptyDirVolumeSource(
                        medium="Memory",
                    )
                )
            )])
            pod_spec.containers.extend([primary_container, secondary_container])
            return pod_spec
        @outputs(out=Types.Integer)
        @python_task
        def my_sub_task(wf_params, out):
            out.set(randint())
        @outputs(out=[Types.Integer])
        @dynamic_sidecar_task(
            pod_spec=generate_pod_spec_for_task(),
            primary_container_name="primary",
        )
        def my_task(wf_params, out):
            out_list = []
            for i in xrange(100):
                out_list.append(my_sub_task().outputs.out)
            out.set(out_list)
    .. note::
        All outputs of a batch task must be a list.  This is because the individual outputs of sub-tasks should be
        appended into a list.  There cannot be aggregation of outputs done in this task.  To accomplish aggregation,
        it is recommended that a python_task take the outputs of this task as input and do the necessary work.
        If a sub-task does not contribute an output, it must be yielded from the task with the `yield` keyword or
        returned from the task in a list.  If this isn't done, the sub-task will not be executed.
    :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed.
    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.
        .. note::
            This argument is required to be a non-empty string if `cache` is True.
    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.sdk.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.
        .. note::
            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.
    :param bool interruptible: [optional] boolean describing if the task is interruptible.
    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.
    :param Text storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for the task to run.  Default is set by platform-level configuration.
        .. note::
            This is currently not supported by the platform.
    :param Text cpu_request: [optional] Kubernetes resource string for lower-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs.
        Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for the task to execute.  Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for the task to run.  This amount is not guaranteed!  If not specified, it is set equal to storage_request.
        .. note::
            This is currently not supported by the platform.
    :param Text cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.
    :param Text gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs. This amount is not
        guaranteed!  If not specified, it is set equal to gpu_request.
    :param Text memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for the task to execute.  This amount is not guaranteed!  If not specified, it is set equal to
        memory_request.
    :param bool cache: [optional] boolean describing if the outputs of this task should be cached and
        re-usable.
    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.
    :param float allowed_failure_ratio: [optional] float value describing the ratio of sub-tasks that may fail before
        the master batch task considers itself a failure.  By default, the value is 0 so if any task fails, the master
        batch task will be marked a failure.  If specified, the value must be between 0 and 1 inclusive.  In the event a
        non-zero value is specified, downstream tasks must be able to accept None values as outputs from individual
        sub-tasks because the output values will be set to None for any sub-task that fails.
    :param int max_concurrency: [optional] integer value describing the maximum number of tasks to run concurrently.
        This is a stand-in pending better concurrency controls for special use-cases.  The existence of this parameter
        is not guaranteed between versions and therefore it is NOT recommended that it be used.
    :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.
    :param k8s.io.api.core.v1.generated_pb2.PodSpec pod_spec: PodSpec to bring up alongside task execution.
    :param Text primary_container_name: primary container to monitor for the duration of the task.
    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided must be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask.  Generally, it should be a
        subclass of flytekit.common.tasks.sidecar_Task.SdkDynamicSidecarTask.  A user can use this parameter to inject bespoke
        logic into the base Flyte programming model.
    :rtype: flytekit.common.tasks.sidecar_Task.SdkDynamicSidecarTask
    """

    def wrapper(fn):
        return (cls or _sdk_sidecar_tasks.SdkDynamicSidecarTask)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.SIDECAR_TASK,
            discovery_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            storage_request=storage_request,
            cpu_request=cpu_request,
            gpu_request=gpu_request,
            memory_request=memory_request,
            storage_limit=storage_limit,
            cpu_limit=cpu_limit,
            gpu_limit=gpu_limit,
            memory_limit=memory_limit,
            discoverable=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            allowed_failure_ratio=allowed_failure_ratio,
            max_concurrency=max_concurrency,
            environment=environment,
            pod_spec=pod_spec,
            primary_container_name=primary_container_name,
        )

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper


def pytorch_task(
    _task_function=None,
    cache_version="",
    retries=0,
    interruptible=False,
    deprecated="",
    cache=False,
    timeout=None,
    workers_count=1,
    per_replica_storage_request="",
    per_replica_cpu_request="",
    per_replica_gpu_request="",
    per_replica_memory_request="",
    per_replica_storage_limit="",
    per_replica_cpu_limit="",
    per_replica_gpu_limit="",
    per_replica_memory_limit="",
    environment=None,
    cls=None,
):
    """
    Decorator to create a Pytorch Task definition. This task will submit PyTorchJob (see https://github.com/kubeflow/pytorch-operator)
        defined by the code within the _task_function to k8s cluster.

    .. code-block:: python

        @inputs(int_list=[Types.Integer])
        @outputs(result=Types.Integer
        @pytorch_task(
            workers_count=2,
            per_replica_cpu_request="500m",
            per_replica_memory_request="4Gi",
            per_replica_memory_limit="8Gi",
            per_replica_gpu_limit="1",
        )
        def my_pytorch_job(wf_params, int_list, result):
            pass

    :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed for wrapped task functions.

    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.

        .. note::

            This argument is required to be a non-empty string if `cache` is True.

    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.sdk.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.

        .. note::

            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.

    :param bool interruptible: [optional] boolean describing if the task is interruptible.

    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.

    :param bool cache: [optional] boolean describing if the outputs of this task should be cached and
        re-usable.

    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.

    :param int workers_count: integer determining the number of worker replicas spawned in the cluster for this job
        (in addition to 1 master).

    :param Text per_replica_storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for each replica spawned for this job (i.e. both for master and workers).  Default is set by platform-level configuration.

        .. note::

            This is currently not supported by the platform.

    :param Text per_replica_cpu_request: [optional] Kubernetes resource string for lower-bound of cores for each replica
        spawned for this job (i.e. both for master and workers).
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text per_replica_gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs for each
        replica spawned for this job (i.e. both for master and workers).
        Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text per_replica_memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for each replica spawned for this job (i.e. both for master and workers).  Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text per_replica_storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for each replica spawned for this job (i.e. both for master and workers).
        This amount is not guaranteed!  If not specified, it is set equal to storage_request.

        .. note::

            This is currently not supported by the platform.

    :param Text per_replica_cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for each replica
        spawned for this job (i.e. both for master and workers).
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.

    :param Text per_replica_gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs for each
        replica spawned for this job (i.e. both for master and workers).
        This amount is not guaranteed!  If not specified, it is set equal to gpu_request.

    :param Text per_replica_memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for each replica spawned for this job (i.e. both for master and workers).
        This amount is not guaranteed!  If not specified, it is set equal to memory_request.

    :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.

    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided must be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask.  A user can use this to
        inject bespoke logic into the base Flyte programming model.

    :rtype: flytekit.common.tasks.sdk_runnable.SdkRunnableTask
    """

    def wrapper(fn):
        return (cls or _sdk_pytorch_tasks.SdkPyTorchTask)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.PYTORCH_TASK,
            discovery_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            discoverable=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            workers_count=workers_count,
            per_replica_storage_request=per_replica_storage_request,
            per_replica_cpu_request=per_replica_cpu_request,
            per_replica_gpu_request=per_replica_gpu_request,
            per_replica_memory_request=per_replica_memory_request,
            per_replica_storage_limit=per_replica_storage_limit,
            per_replica_cpu_limit=per_replica_cpu_limit,
            per_replica_gpu_limit=per_replica_gpu_limit,
            per_replica_memory_limit=per_replica_memory_limit,
            environment=environment or {},
        )

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper


def tensorflow_task(
    _task_function=None,
    cache_version="",
    retries=0,
    interruptible=False,
    deprecated="",
    cache=False,
    timeout=None,
    workers_count=1,
    ps_replicas_count=None,
    chief_replicas_count=None,
    per_replica_storage_request="",
    per_replica_cpu_request="",
    per_replica_gpu_request="",
    per_replica_memory_request="",
    per_replica_storage_limit="",
    per_replica_cpu_limit="",
    per_replica_gpu_limit="",
    per_replica_memory_limit="",
    environment=None,
    cls=None,
):
    """
    Decorator to create a Tensorflow Task definition. This task will submit TFJob (see https://github.com/kubeflow/tf-operator)
        defined by the code within the _task_function to k8s cluster.

    .. code-block:: python

        @inputs(int_list=[Types.Integer])
        @outputs(result=Types.Integer
        @tensorflow_task(
            workers_count=2,
            ps_replicas_count=1,
            chief_replicas_count=1,
            per_replica_cpu_request="500m",
            per_replica_memory_request="4Gi",
            per_replica_memory_limit="8Gi",
            per_replica_gpu_limit="1",
        )
        def my_tensorflow_job(wf_params, int_list, result):
            pass

    :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed for wrapped task functions.

    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.

        .. note::

            This argument is required to be a non-empty string if `cache` is True.

    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.sdk.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.

        .. note::

            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.

    :param bool interruptible: [optional] boolean describing if the task is interruptible.

    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.

    :param bool cache: [optional] boolean describing if the outputs of this task should be cached and
        re-usable.

    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.

    :param int workers_count: integer determining the number of worker replicas spawned in the cluster for this job

    :param int ps_replicas_count: integer determining the number of parameter server replicas spawned in the cluster for this job

    :param int chief_replicas_count: integer determining the number of chief server replicas spawned in the cluster for this job

    :param Text per_replica_storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for each replica spawned for this job (i.e. both for parameter, chief server and workers).  Default is set by platform-level configuration.

        .. note::

            This is currently not supported by the platform.

    :param Text per_replica_cpu_request: [optional] Kubernetes resource string for lower-bound of cores for each replica
        spawned for this job (i.e. both for parameter, chief server and workers).
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text per_replica_gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs for each
        replica spawned for this job (i.e. both for parameter, chief server and workers).
        Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text per_replica_memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for each replica spawned for this job (i.e. both for parameter, chief server and workers).  Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text per_replica_storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for each replica spawned for this job (i.e. both for parameter, chief server and workers).
        This amount is not guaranteed!  If not specified, it is set equal to storage_request.

        .. note::

            This is currently not supported by the platform.

    :param Text per_replica_cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for each replica
        spawned for this job (i.e. both for parameter, chief server and workers).
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.

    :param Text per_replica_gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs for each
        replica spawned for this job (i.e. both for parameter, chief server and workers).
        This amount is not guaranteed!  If not specified, it is set equal to gpu_request.

    :param Text per_replica_memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for each replica spawned for this job (i.e. both for parameter, chief server and workers).
        This amount is not guaranteed!  If not specified, it is set equal to memory_request.

    :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.

    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided must be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask.  A user can use this to
        inject bespoke logic into the base Flyte programming model.

    :rtype: flytekit.common.tasks.sdk_runnable.SdkRunnableTask
    """

    def wrapper(fn):
        return (cls or _sdk_tensorflow_tasks.SdkTensorFlowTask)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.TENSORFLOW_TASK,
            cache_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            cache=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            workers_count=workers_count,
            ps_replicas_count=ps_replicas_count,
            chief_replicas_count=chief_replicas_count,
            per_replica_storage_request=per_replica_storage_request,
            per_replica_cpu_request=per_replica_cpu_request,
            per_replica_gpu_request=per_replica_gpu_request,
            per_replica_memory_request=per_replica_memory_request,
            per_replica_storage_limit=per_replica_storage_limit,
            per_replica_cpu_limit=per_replica_cpu_limit,
            per_replica_gpu_limit=per_replica_gpu_limit,
            per_replica_memory_limit=per_replica_memory_limit,
            environment=environment or {},
        )

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper



from flyteidl.admin import launch_plan_pb2 as _launch_plan

from flytekit.models import common as _common
from flytekit.models import interface as _interface
from flytekit.models import literals as _literals
from flytekit.models import schedule as _schedule
from flytekit.models.core import identifier as _identifier


class LaunchPlanMetadata(_common.FlyteIdlEntity):
    def __init__(self, schedule, notifications):
        """

        :param flytekit.models.schedule.Schedule schedule: Schedule to execute the Launch Plan
        :param list[flytekit.models.common.Notification] notifications: List of notifications based on
            execution status transitions
        """
        self._schedule = schedule
        self._notifications = notifications

    @property
    def schedule(self):
        """
        Schedule to execute the Launch Plan
        :rtype: flytekit.models.schedule.Schedule
        """
        return self._schedule

    @property
    def notifications(self):
        """
        List of notifications based on Execution status transitions
        :rtype: list[flytekit.models.common.Notification]
        """
        return self._notifications

    def to_flyte_idl(self):
        """
        List of notifications based on Execution status transitions
        :rtype: flyteidl.admin.launch_plan_pb2.LaunchPlanMetadata
        """
        return _launch_plan.LaunchPlanMetadata(
            schedule=self.schedule.to_flyte_idl() if self.schedule is not None else None,
            notifications=[n.to_flyte_idl() for n in self.notifications],
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.launch_plan_pb2.LaunchPlanMetadata pb2_object:
        :rtype: LaunchPlanMetadata
        """
        return cls(
            schedule=_schedule.Schedule.from_flyte_idl(pb2_object.schedule)
            if pb2_object.HasField("schedule")
            else None,
            notifications=[_common.Notification.from_flyte_idl(n) for n in pb2_object.notifications],
        )


class Auth(_common.FlyteIdlEntity):
    def __init__(self, assumable_iam_role=None, kubernetes_service_account=None):
        """
        DEPRECATED. Do not use. Use flytekit.models.common.AuthRole instead
        At most one of assumable_iam_role or kubernetes_service_account can be set.
        :param Text assumable_iam_role: IAM identity with set permissions policies.
        :param Text kubernetes_service_account: Provides an identity for workflow execution resources. Flyte deployment
            administrators are responsible for handling permissions as they relate to the service account.
        """
        if assumable_iam_role and kubernetes_service_account:
            raise ValueError("Only one of assumable_iam_role or kubernetes_service_account can be set")
        self._assumable_iam_role = assumable_iam_role
        self._kubernetes_service_account = kubernetes_service_account

    @property
    def assumable_iam_role(self):
        """
        The IAM role to execute the workflow with
        :rtype: Text
        """
        return self._assumable_iam_role

    @property
    def kubernetes_service_account(self):
        """
        The kubernetes service account to execute the workflow with
        :rtype: Text
        """
        return self._kubernetes_service_account

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.launch_plan_pb2.Auth
        """
        return _launch_plan.Auth(
            assumable_iam_role=self.assumable_iam_role if self.assumable_iam_role else None,
            kubernetes_service_account=self.kubernetes_service_account if self.kubernetes_service_account else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.launch_plan_pb2.Auth pb2_object:
        :rtype: Auth
        """
        return cls(
            assumable_iam_role=pb2_object.assumable_iam_role if pb2_object.HasField("assumable_iam_role") else None,
            kubernetes_service_account=pb2_object.kubernetes_service_account
            if pb2_object.HasField("kubernetes_service_account")
            else None,
        )


class LaunchPlanSpec(_common.FlyteIdlEntity):
    def __init__(
        self,
        workflow_id,
        entity_metadata,
        default_inputs,
        fixed_inputs,
        labels,
        annotations,
        auth_role,
        raw_output_data_config,
    ):
        """
        The spec for a Launch Plan.

        :param flytekit.models.core.identifier.Identifier workflow_id: Unique identifier for the workflow in question
        :param LaunchPlanMetadata entity_metadata: Metadata
        :param flytekit.models.interface.ParameterMap default_inputs: Input values to be passed for the execution
        :param flytekit.models.literals.LiteralMap fixed_inputs: Fixed, non-overridable inputs for the Launch Plan
        :param flyteidl.admin.common_pb2.Labels:
            Any custom kubernetes labels to apply to workflows executed by this launch plan.
        :param flyteidl.admin.common_pb2.Annotations annotations:
            Any custom kubernetes annotations to apply to workflows executed by this launch plan.
        :param flytekit.models.common.Auth auth_role: The auth method with which to execute the workflow.
        :param flytekit.models.common.RawOutputDataConfig raw_output_data_config: Value for where to store offloaded
            data like Blobs and Schemas.
        """
        self._workflow_id = workflow_id
        self._entity_metadata = entity_metadata
        self._default_inputs = default_inputs
        self._fixed_inputs = fixed_inputs
        self._labels = labels
        self._annotations = annotations
        self._auth_role = auth_role
        self._raw_output_data_config = raw_output_data_config

    @property
    def workflow_id(self):
        """
        Unique identifier for the workflow in question
        :rtype: flytekit.models.core.identifier.Identifier
        """
        return self._workflow_id

    @property
    def entity_metadata(self):
        """
        :rtype: LaunchPlanMetadata
        """
        return self._entity_metadata

    @property
    def default_inputs(self):
        """
        Input values to be passed for the execution
        :rtype: flytekit.models.interface.ParameterMap
        """
        return self._default_inputs

    @property
    def fixed_inputs(self):
        """
        Fixed, non-overridable inputs for the Launch Plan
        :rtype: flytekit.models.literals.LiteralMap
        """
        return self._fixed_inputs

    @property
    def labels(self):
        """
        The labels to execute the workflow with
        :rtype: flytekit.models.common.Labels
        """
        return self._labels

    @property
    def annotations(self):
        """
        The annotations to execute the workflow with
        :rtype: flytekit.models.common.Annotations
        """
        return self._annotations

    @property
    def auth_role(self):
        """
        The authorization method with which to execute the workflow.
        :rtype: flytekit.models.common.Auth
        """
        return self._auth_role

    @property
    def raw_output_data_config(self):
        """
        Where to store offloaded data like Blobs and Schemas
        :rtype: flytekit.models.common.RawOutputDataConfig
        """
        return self._raw_output_data_config

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.launch_plan_pb2.LaunchPlanSpec
        """
        return _launch_plan.LaunchPlanSpec(
            workflow_id=self.workflow_id.to_flyte_idl(),
            entity_metadata=self.entity_metadata.to_flyte_idl(),
            default_inputs=self.default_inputs.to_flyte_idl(),
            fixed_inputs=self.fixed_inputs.to_flyte_idl(),
            labels=self.labels.to_flyte_idl(),
            annotations=self.annotations.to_flyte_idl(),
            auth_role=self.auth_role.to_flyte_idl(),
            raw_output_data_config=self.raw_output_data_config.to_flyte_idl(),
        )

    @classmethod
    def from_flyte_idl(cls, pb2):
        """
        :param flyteidl.admin.launch_plan_pb2.LaunchPlanSpec pb2:
        :rtype: LaunchPlanSpec
        """
        auth_role = None
        # First check the newer field, auth_role.
        if pb2.auth_role is not None and (pb2.auth_role.assumable_iam_role or pb2.auth_role.kubernetes_service_account):
            auth_role = _common.AuthRole.from_flyte_idl(pb2.auth_role)
        # Fallback to the deprecated field.
        elif pb2.auth is not None:
            if pb2.auth.assumable_iam_role:
                auth_role = _common.AuthRole(assumable_iam_role=pb2.auth.assumable_iam_role)
            else:
                auth_role = _common.AuthRole(assumable_iam_role=pb2.auth.kubernetes_service_account)

        return cls(
            workflow_id=_identifier.Identifier.from_flyte_idl(pb2.workflow_id),
            entity_metadata=LaunchPlanMetadata.from_flyte_idl(pb2.entity_metadata),
            default_inputs=_interface.ParameterMap.from_flyte_idl(pb2.default_inputs),
            fixed_inputs=_literals.LiteralMap.from_flyte_idl(pb2.fixed_inputs),
            labels=_common.Labels.from_flyte_idl(pb2.labels),
            annotations=_common.Annotations.from_flyte_idl(pb2.annotations),
            auth_role=auth_role,
            raw_output_data_config=_common.RawOutputDataConfig.from_flyte_idl(pb2.raw_output_data_config),
        )


class LaunchPlanState(object):
    INACTIVE = _launch_plan.INACTIVE
    ACTIVE = _launch_plan.ACTIVE

    @classmethod
    def enum_to_string(cls, val):
        """
        :param int val:
        :rtype: Text
        """
        if val == cls.INACTIVE:
            return "INACTIVE"
        elif val == cls.ACTIVE:
            return "ACTIVE"
        else:
            return "<UNKNOWN>"


class LaunchPlanClosure(_common.FlyteIdlEntity):
    def __init__(self, state, expected_inputs, expected_outputs):
        """
        :param LaunchPlanState state: Indicate the Launch plan phase
        :param flytekit.models.interface.ParameterMap expected_inputs: Indicates the set of inputs to execute
            the Launch plan
        :param flytekit.models.interface.VariableMap expected_outputs: Indicates the set of outputs from the Launch plan
        """
        self._state = state
        self._expected_inputs = expected_inputs
        self._expected_outputs = expected_outputs

    @property
    def state(self):
        """
        :rtype: LaunchPlanState
        """
        return self._state

    @property
    def expected_inputs(self):
        """
        :rtype: flytekit.models.interface.ParameterMap
        """
        return self._expected_inputs

    @property
    def expected_outputs(self):
        """
        :rtype: flytekit.models.interface.VariableMap
        """
        return self._expected_outputs

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.launch_plan_pb2.LaunchPlanClosure
        """
        return _launch_plan.LaunchPlanClosure(
            state=self.state,
            expected_inputs=self.expected_inputs.to_flyte_idl(),
            expected_outputs=self.expected_outputs.to_flyte_idl(),
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.launch_plan_pb2.LaunchPlanClosure pb2_object:
        :rtype: LaunchPlanClosure
        """
        return cls(
            pb2_object.state,
            _interface.ParameterMap.from_flyte_idl(pb2_object.expected_inputs),
            _interface.VariableMap.from_flyte_idl(pb2_object.expected_outputs),
        )


class LaunchPlan(_common.FlyteIdlEntity):
    def __init__(self, id, spec, closure):
        """
        :param flytekit.models.core.identifier.Identifier id:
        :param LaunchPlanSpec spec:
        :param LaunchPlanClosure closure:
        """
        self._id = id
        self._spec = spec
        self._closure = closure

    @property
    def id(self):
        """
        :rtype: flytekit.models.core.identifier.Identifier
        """
        return self._id

    @property
    def spec(self):
        """
        :rtype: LaunchPlanSpec
        """
        return self._spec

    @property
    def closure(self):
        """
        :rtype: LaunchPlanClosure
        """
        return self._closure

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.launch_plan_pb2.LaunchPlan
        """
        identifier = (
            self.id
            if self.id is not None
            else _identifier.Identifier(_identifier.ResourceType.LAUNCH_PLAN, None, None, None, None)
        )
        return _launch_plan.LaunchPlan(
            id=identifier.to_flyte_idl(),
            spec=self.spec.to_flyte_idl(),
            closure=self.closure.to_flyte_idl(),
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.launch_plan_pb2.LaunchPlan pb2_object:
        :rtype: LaunchPlan
        """
        return cls(
            id=_identifier.Identifier.from_flyte_idl(pb2_object.id),
            spec=LaunchPlanSpec.from_flyte_idl(pb2_object.spec),
            closure=LaunchPlanClosure.from_flyte_idl(pb2_object.closure),
        )



from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.models import common as _common_model
from flytekit.models.core import execution as _execution_model


class Notification(_common_model.Notification, metaclass=_sdk_bases.ExtendedSdkType):

    VALID_PHASES = {
        _execution_model.WorkflowExecutionPhase.ABORTED,
        _execution_model.WorkflowExecutionPhase.FAILED,
        _execution_model.WorkflowExecutionPhase.SUCCEEDED,
        _execution_model.WorkflowExecutionPhase.TIMED_OUT,
    }

    def __init__(self, phases, email=None, pager_duty=None, slack=None):
        """
        :param list[int] phases: A required list of phases for which to fire the event.  Events can only be fired for
            terminal phases.  Phases should be as defined in: flytekit.models.core.execution.WorkflowExecutionPhase
        """
        self._validate_phases(phases)
        super(Notification, self).__init__(phases, email=email, pager_duty=pager_duty, slack=slack)

    def _validate_phases(self, phases):
        """
        :param list[int] phases:
        """
        if len(phases) == 0:
            raise _user_exceptions.FlyteAssertion("You must specify at least one phase for a notification.")
        for phase in phases:
            if phase not in self.VALID_PHASES:
                raise _user_exceptions.FlyteValueException(
                    phase,
                    self.VALID_PHASES,
                    additional_message="Notifications can only be specified on terminal states.",
                )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.admin.common_pb2.Notification p:
        :rtype: Notification
        """
        if p.HasField("email"):
            return cls(p.phases, p.email.recipients_email)
        elif p.HasField("pager_duty"):
            return cls(p.phases, p.pager_duty.recipients_email)
        else:
            return cls(p.phases, p.slack.recipients_email)


class PagerDuty(Notification):
    def __init__(self, phases, recipients_email):
        """
        :param list[Text] recipients_email: A required non-empty list of recipients for the notification.
        """
        super(PagerDuty, self).__init__(phases, pager_duty=_common_model.PagerDutyNotification(recipients_email))

    @classmethod
    def promote_from_model(cls, base_model):
        """
        :param flytekit.models.common.Notification base_model:
        :rtype: Notification
        """
        return cls(base_model.phases, base_model.pager_duty.recipients_email)


class Email(Notification):
    def __init__(self, phases, recipients_email):
        """
        :param list[Text] recipients_email: A required non-empty list of recipients for the notification.
        :param list[int] phases: A required list of phases for which to fire the event.  Events can only be fired for
            terminal phases.  Phases should be as defined in: flytekit.models.core.execution.WorkflowExecutionPhase
        """
        super(Email, self).__init__(phases, email=_common_model.EmailNotification(recipients_email))

    @classmethod
    def promote_from_model(cls, base_model):
        """
        :param flytekit.models.common.Notification base_model:
        :rtype: Notification
        """
        return cls(base_model.phases, base_model.email.recipients_email)


class Slack(Notification):
    def __init__(self, phases, recipients_email):
        """
        :param list[Text] recipients_email: A required non-empty list of recipients for the notification.
        :param list[int] phases: A required list of phases for which to fire the event.  Events can only be fired for
            terminal phases.  Phases should be as defined in: flytekit.models.core.execution.WorkflowExecutionPhase
        """
        super(Slack, self).__init__(phases, slack=_common_model.SlackNotification(recipients_email))

    @classmethod
    def promote_from_model(cls, base_model):
        """
        :param flytekit.models.common.Notification base_model:
        :rtype: Notification
        """
        return cls(base_model.phases, base_model.slack.recipients_email)



"""
Notifications are primarily used when defining Launch Plans (also can be used when launching executions) and will trigger
the Flyte platform to send emails when a workflow run reaches certain stages (fails or suceeds, etc.).

.. note::

    Notifications require some setup and configuration on the Flyte platform side. Please contact your Flyte platform
    admins to get this feature enabled. See :std:doc:`flyte:howto/notifications`

Each notification type takes a list of :py:class:`flytekit.models.core.execution.WorkflowExecutionPhase` and a list of
emails. Even though there are different notification classes in this module, they all just send email. The differentiation
offers semantic meaning to the end-user but do not functionally behave differently. Successful integration with Slack
and Pagerduty is incumbent on those email API being set-up correctly.

.. autoclass:: flytekit.core.notification.Notification

"""
from typing import List

from flytekit.models import common as _common_model
from flytekit.models.core import execution as _execution_model


# Duplicates flytekit.common.notifications.Notification to avoid using the ExtendedSdkType metaclass.
class Notification(_common_model.Notification):

    VALID_PHASES = {
        _execution_model.WorkflowExecutionPhase.ABORTED,
        _execution_model.WorkflowExecutionPhase.FAILED,
        _execution_model.WorkflowExecutionPhase.SUCCEEDED,
        _execution_model.WorkflowExecutionPhase.TIMED_OUT,
    }

    def __init__(
        self,
        phases: List[int],
        email: _common_model.EmailNotification = None,
        pager_duty: _common_model.PagerDutyNotification = None,
        slack: _common_model.SlackNotification = None,
    ):
        """
        :param list[int] phases: A required list of phases for which to fire the event.  Events can only be fired for
            terminal phases.  Phases should be as defined in: flytekit.models.core.execution.WorkflowExecutionPhase
        """
        self._validate_phases(phases)
        super(Notification, self).__init__(phases, email=email, pager_duty=pager_duty, slack=slack)

    def _validate_phases(self, phases: List[int]):
        """
        :param list[int] phases:
        """
        if len(phases) == 0:
            raise AssertionError("You must specify at least one phase for a notification.")
        for phase in phases:
            if phase not in self.VALID_PHASES:
                raise AssertionError(f"Invalid phase: {phase}. only terminal states are permitted for notifications")


class PagerDuty(Notification):
    """
    This notification should be used when sending emails to the PagerDuty service.
    """

    def __init__(self, phases: List[int], recipients_email: List[str]):
        """
        :param list[int] phases: A required list of phases for which to fire the event.  Events can only be fired for
            terminal phases.  Phases should be as defined in: flytekit.models.core.execution.WorkflowExecutionPhase
        :param list[str] recipients_email: A required non-empty list of recipients for the notification.
        """
        super(PagerDuty, self).__init__(phases, pager_duty=_common_model.PagerDutyNotification(recipients_email))


class Email(Notification):
    """
    This notification should be used when sending regular emails to people.
    """

    def __init__(self, phases: List[int], recipients_email: List[str]):
        """
        :param list[int] phases: A required list of phases for which to fire the event.  Events can only be fired for
            terminal phases. Phases should be as defined in: :py:class:`flytekit.models.core.execution.WorkflowExecutionPhase`
        :param list[str] recipients_email: A required non-empty list of recipients for the notification.
        """
        super(Email, self).__init__(phases, email=_common_model.EmailNotification(recipients_email))


class Slack(Notification):
    """
    This notification should be used when sending emails to the Slack.
    """

    def __init__(self, phases: List[int], recipients_email: List[str]):
        """
        :param list[int] phases: A required list of phases for which to fire the event.  Events can only be fired for
            terminal phases.  Phases should be as defined in: flytekit.models.core.execution.WorkflowExecutionPhase
        :param list[str] recipients_email: A required non-empty list of recipients for the notification.
        """
        super(Slack, self).__init__(phases, slack=_common_model.SlackNotification(recipients_email))



from flytekit.common import constants as _constants
from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import helpers as _type_helpers
from flytekit.models import interface as _interface_models
from flytekit.models import types as _type_models


class Input(_interface_models.Parameter, metaclass=_sdk_bases.ExtendedSdkType):
    def __init__(self, name, sdk_type, help=None, **kwargs):
        """
        :param Text name:
        :param flytekit.common.types.base_sdk_types.FlyteSdkType sdk_type: This is the SDK type necessary to create an
            input to this workflow.
        :param Text help: An optional help string to describe the input to users.
        :param bool required: If set to True, default must be None
        :param T default:  If this is not a required input, the value will default to this value.
        """
        param_default = None
        if "required" not in kwargs and "default" not in kwargs:
            # Neither required or default is set so assume required
            required = True
            default = None
        elif kwargs.get("required", False) and "default" in kwargs:
            # Required cannot be set to True and have a default specified
            raise _user_exceptions.FlyteAssertion("Default cannot be set when required is True")
        elif "default" in kwargs:
            # If default is specified, then required must be false and the value is whatever is specified
            required = None
            default = kwargs["default"]
            param_default = sdk_type.from_python_std(default)
        else:
            # If no default is set, but required is set, then the behavior is determined by required == True or False
            default = None
            required = kwargs["required"]
            if not required:
                # If required == False, we assume default to be None
                param_default = sdk_type.from_python_std(default)
                required = None

        self._sdk_required = required or False
        self._sdk_default = default
        self._help = help
        self._sdk_type = sdk_type
        self._promise = _type_models.OutputReference(_constants.GLOBAL_INPUT_NODE_ID, name)
        self._name = name
        super(Input, self).__init__(
            _interface_models.Variable(type=sdk_type.to_flyte_literal_type(), description=help or ""),
            required=required,
            default=param_default,
        )

    def rename_and_return_reference(self, new_name):
        self._promise._var = new_name
        return self

    @property
    def name(self):
        """
        :rtype: Text
        """
        return self._promise.var

    @property
    def promise(self):
        """
        :rtype: flytekit.models.types.OutputReference
        """
        return self._promise

    @property
    def sdk_required(self):
        """
        :rtype: bool
        """
        return self._sdk_required

    @property
    def sdk_default(self):
        """
        :rtype: T
        """
        return self._sdk_default

    @property
    def help(self):
        """
        :rtype: Text
        """
        return self._help

    @property
    def sdk_type(self):
        """
        :rtype: flytekit.common.types.base_sdk_types.FlyteSdkType
        """
        return self._sdk_type

    def __repr__(self):
        return "Input({}, {}, required={}, help={})".format(self.name, self.sdk_type, self.required, self.help)

    @classmethod
    def promote_from_model(cls, model):
        """
        :param flytekit.models.interface.Parameter model:
        :rtype: Input
        """
        sdk_type = _type_helpers.get_sdk_type_from_literal_type(model.var.type)

        if model.default is not None:
            default_value = sdk_type.from_flyte_idl(model.default.to_flyte_idl()).to_python_std()
            return cls(
                "",
                sdk_type,
                help=model.var.description,
                required=False,
                default=default_value,
            )
        else:
            return cls("", sdk_type, help=model.var.description, required=True)


class NodeOutput(_type_models.OutputReference, metaclass=_sdk_bases.ExtendedSdkType):
    def __init__(self, sdk_node, sdk_type, var):
        """
        :param sdk_node:
        :param sdk_type: deprecated in mypy flytekit.
        :param var:
        """
        self._node = sdk_node
        self._type = sdk_type
        super(NodeOutput, self).__init__(self._node.id, var)

    @property
    def node_id(self):
        """
        Override the underlying node_id property to refer to SdkNode.
        :rtype: Text
        """
        return self.sdk_node.id

    @classmethod
    def promote_from_model(cls, model):
        """
        :param flytekit.models.types.OutputReference model:
        :rtype: NodeOutput
        """
        raise _user_exceptions.FlyteAssertion(
            "A NodeOutput cannot be promoted from a protobuf because it must be "
            "contextualized by an existing SdkNode."
        )

    @property
    def sdk_node(self):
        """
        :rtype: flytekit.common.nodes.SdkNode
        """
        return self._node

    @property
    def sdk_type(self):
        """
        :rtype: flytekit.common.types.base_sdk_types.FlyteSdkType
        """
        return self._type

    def __repr__(self):
        s = f"NodeOutput({self.sdk_node if self.sdk_node.id is not None else None}:{self.var})"
        return s



import os as _os
import shutil as _shutil
import sys as _sys
import uuid as _uuid

import six as _six

from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common import utils as _utils
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.interfaces.data import data_proxy as _data_proxy
from flytekit.models import literals as _literal_models
from flytekit.models.core import types as _core_types


class Blob(_literal_models.Blob, metaclass=_sdk_bases.ExtendedSdkType):
    def __init__(self, remote_path, mode="rb", format=None):
        """
        :param Text remote_path: Path to location where the Blob should be synced to.
        :param Text mode: File access mode.  'a' and '+' are forbidden.  A blob can only be written or read at a time.
        :param Text format: Format
        """
        if "+" in mode or "a" in mode or ("w" in mode and "r" in mode):
            raise _user_exceptions.FlyteAssertion("A blob cannot be read and written at the same time")
        self._mode = mode
        self._local_path = None
        self._file = None
        super(Blob, self).__init__(
            _literal_models.BlobMetadata(
                type=_core_types.BlobType(format or "", _core_types.BlobType.BlobDimensionality.SINGLE)
            ),
            remote_path,
        )

    @classmethod
    @_exception_scopes.system_entry_point
    def from_python_std(cls, t_value, mode="wb", format=None):
        """
        :param T t_value:
        :param Text mode: File access mode.  'a' and '+' are forbidden.  A blob can only be written or read at a time.
        :param Text format:
        :rtype: Blob
        """
        if isinstance(t_value, (_six.text_type, str)):
            if _os.path.isfile(t_value):
                blob = cls.create_at_any_location(mode=mode, format=format)
                blob._local_path = t_value
                blob.upload()
            else:
                blob = cls.create_at_known_location(t_value, mode=mode, format=format)
            return blob
        elif isinstance(t_value, cls):
            return t_value
        else:
            raise _user_exceptions.FlyteTypeException(
                type(t_value),
                {_six.text_type, str, Blob},
                received_value=t_value,
                additional_msg="Unable to create Blob from user-provided value.",
            )

    @classmethod
    @_exception_scopes.system_entry_point
    def from_string(cls, t_value, mode="wb", format=None):
        """
        :param T t_value:
        :param Text mode: Read or write mode of the object.
        :param Text format:
        :rtype: Blob
        """
        return cls.create_at_known_location(t_value, mode=mode, format=format)

    @classmethod
    @_exception_scopes.system_entry_point
    def create_at_known_location(cls, known_remote_location, mode="wb", format=None):
        """
        :param Text known_remote_location: The location to which to write the object.  Usually an s3 path.
        :param Text mode:
        :param Text format:
        :rtype: Blob
        """
        return cls(known_remote_location, mode=mode, format=format)

    @classmethod
    @_exception_scopes.system_entry_point
    def create_at_any_location(cls, mode="wb", format=None):
        """
        :param Text mode:
        :param Text format:
        :rtype: Blob
        """
        return cls.create_at_known_location(_data_proxy.Data.get_remote_path(), mode=mode, format=format)

    @classmethod
    @_exception_scopes.system_entry_point
    def fetch(cls, remote_path, local_path=None, overwrite=False, mode="rb", format=None):
        """
        :param Text remote_path: The location from which to fetch the object. Usually an s3 path.
        :param Text local_path: [Optional] A local path to which to download the object. If specified, the object
            will not be managed and might not be cleaned up by the system upon exiting the context.
        :param bool overwrite: If True, objects will be overwritten at the provided local_path in order to fetch this
            object.  Default is False.
        :param Text mode: Read or write mode of the object.
        :param Text format: Format the object is in.
        :rtype: Blob
        """
        blob = cls(remote_path, mode=mode, format=format)
        blob.download(local_path=local_path, overwrite=overwrite)
        return blob

    @classmethod
    def promote_from_model(cls, model, mode="rb"):
        """
        :param flytekit.models.literals.Blob model:
        :param Text mode: Read or write mode of the object.
        :rtype: Blob
        """
        return cls(model.uri, format=model.metadata.type.format, mode=mode)

    @property
    def local_path(self):
        """
        Local filesystem path where the file was downloaded
        :rtype: Text
        """
        return self._local_path

    @property
    def remote_location(self):
        """
        Path to where this Blob will be synced.
        :rtype: Text
        """
        return self.uri

    @property
    def mode(self):
        """
        The mode string the Blob is associated with.
        :rtype: Text
        """
        return self._mode

    @_exception_scopes.system_entry_point
    def __enter__(self):
        """
        :rtype: typing.BinaryIO
        """
        if self._file is not None:
            raise _user_exceptions.FlyteAssertion("Only one reference can be open to a blob at a time.")

        if self.local_path is None:
            if "r" in self.mode:
                self.download()
            elif "w" in self.mode:
                self._generate_local_path()

        self._file = open(self.local_path, self.mode)
        return self._file

    @_exception_scopes.system_entry_point
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._file is not None and not self._file.closed:
            self._file.close()
            self._file = None
            if "w" in self.mode:
                self.upload()
        return False

    def _generate_local_path(self):
        if _data_proxy.LocalWorkingDirectoryContext.get() is None:
            raise _user_exceptions.FlyteAssertion(
                "No temporary file system is present.  Either call this method from within the "
                "context of a task or surround with a 'with LocalTestFileSystem():' block.  Or "
                "specify a path when calling this function.  Note: Cleanup is not automatic when a "
                "path is specified."
            )
        self._local_path = _data_proxy.LocalWorkingDirectoryContext.get().get_named_tempfile(_uuid.uuid4().hex)

    @_exception_scopes.system_entry_point
    def download(self, local_path=None, overwrite=False):
        """
        Alternate method, rather than the context manager interface to download the binary file to the local disk.
        :param Text local_path: [Optional] If provided, the blob will be downloaded to this path.  This will make the
            resulting file object unmanaged and it will not be cleaned up by the system upon exiting the context.
        :param bool overwrite: If true and local_path is specified, we will download the blob and
            overwrite an existing file at that location.  Default is False.
        """
        if "r" not in self._mode:
            raise _user_exceptions.FlyteAssertion("Cannot download a write-only blob!")

        if local_path:
            self._local_path = local_path

        if not self.local_path:
            self._generate_local_path()

        if overwrite or not _os.path.exists(self.local_path):
            # TODO: Introduce system logging
            # logging.info("Getting {} -> {}".format(self.remote_location, self.local_path))
            _data_proxy.Data.get_data(self.remote_location, self.local_path, is_multipart=False)
        else:
            raise _user_exceptions.FlyteAssertion(
                "Cannot download blob to a location that already exists when overwrite is not set to True.  "
                "Attempted download from {} -> {}".format(self.remote_location, self.local_path)
            )

    @_exception_scopes.system_entry_point
    def upload(self):
        """
        Upload the blob to the remote location
        """
        if "w" not in self.mode:
            raise _user_exceptions.FlyteAssertion("Cannot upload a read-only blob!")

        elif not self.local_path:
            raise _user_exceptions.FlyteAssertion(
                "The Blob is not currently backed by a local file and therefore "
                "cannot be uploaded.  Please write to this Blob before attempting "
                "an upload."
            )
        else:
            # TODO: Introduce system logging
            # logging.info("Putting {} -> {}".format(self.local_path, self.remote_location))
            _data_proxy.Data.put_data(self.local_path, self.remote_location, is_multipart=False)


class MultiPartBlob(_literal_models.Blob, metaclass=_sdk_bases.ExtendedSdkType):
    def __init__(self, remote_path, mode="rb", format=None):
        """
        :param Text remote_path: Path to location where the Blob should be synced to.
        :param Text mode: File access mode.  'a' and '+' are forbidden.  A blob can only be written or read at a time.
        :param Text format: Format of underlying blob pieces.
        """
        remote_path = remote_path.strip().rstrip("/") + "/"
        super(MultiPartBlob, self).__init__(
            _literal_models.BlobMetadata(
                type=_core_types.BlobType(format or "", _core_types.BlobType.BlobDimensionality.MULTIPART)
            ),
            remote_path,
        )
        self._is_managed = False
        self._blobs = []
        self._directory = None
        self._mode = mode

    @classmethod
    def promote_from_model(cls, model, mode="rb"):
        """
        :param flytekit.models.literals.Blob model:
        :param Text mode: File access mode.  'a' and '+' are forbidden.  A blob can only be written or read at a time.
        :rtype: Blob
        """
        return cls(model.uri, format=model.metadata.type.format, mode=mode)

    @classmethod
    @_exception_scopes.system_entry_point
    def create_at_known_location(cls, known_remote_location, mode="wb", format=None):
        """
        :param Text known_remote_location: The location to which to write the object.  Usually an s3 path.
        :param Text mode:
        :param Text format:
        :rtype: MultiPartBlob
        """
        return cls(known_remote_location, mode=mode, format=format)

    @classmethod
    @_exception_scopes.system_entry_point
    def create_at_any_location(cls, mode="wb", format=None):
        """
        :param Text mode:
        :param Text format:
        :rtype: MultiPartBlob
        """
        return cls.create_at_known_location(_data_proxy.Data.get_remote_path(), mode=mode, format=format)

    @classmethod
    @_exception_scopes.system_entry_point
    def fetch(cls, remote_path, local_path=None, overwrite=False, mode="rb", format=None):
        """
        :param Text remote_path: The location from which to fetch the object. Usually an s3 path.
        :param Text local_path: [Optional] A local path to which to download the object. If specified, the object
            will not be managed and might not be cleaned up by the system upon exiting the context.
        :param bool overwrite: If True, objects will be overwritten at the provided local_path in order to fetch this
            object.  Default is False.
        :param Text mode: Read or write mode of the object.
        :param Text format: Format the object is in.
        :rtype: MultiPartBlob
        """
        blob = cls(remote_path, mode=mode, format=format)
        blob.download(local_path=local_path, overwrite=overwrite)
        return blob

    @classmethod
    @_exception_scopes.system_entry_point
    def from_python_std(cls, t_value, mode="wb", format=None):
        """
        :param T t_value:
        :param Text mode: Read or write mode of the object.
        :param Text format:
        :rtype: MultiPartBlob
        """
        if isinstance(t_value, (str, _six.text_type)):
            if _os.path.isdir(t_value):
                # TODO: Infer format
                blob = cls.create_at_any_location(mode=mode, format=format)
                blob._directory = _utils.Directory(t_value)
                blob.upload()
            else:
                blob = cls.create_at_known_location(t_value, mode=mode, format=format)
            return blob
        elif isinstance(t_value, cls):
            return t_value
        else:
            raise _user_exceptions.FlyteTypeException(
                type(t_value),
                {str, _six.text_type, MultiPartBlob},
                received_value=t_value,
                additional_msg="Unable to create Blob from user-provided value.",
            )

    @classmethod
    @_exception_scopes.system_entry_point
    def from_string(cls, t_value, mode="wb", format=None):
        """
        :param T t_value:
        :param Text mode: Read or write mode of the object.
        :param Text format:
        :rtype: MultiPartBlob
        """
        return cls.create_at_known_location(t_value, mode=mode, format=format)

    @_exception_scopes.system_entry_point
    def __enter__(self):
        """
        :rtype: list[typing.BinaryIO]
        """
        if "r" not in self.mode:
            raise _user_exceptions.FlyteAssertion("Do not enter context to write to directory. Call create_piece")

        try:
            if not self._directory:
                if _data_proxy.LocalWorkingDirectoryContext.get() is None:
                    raise _user_exceptions.FlyteAssertion(
                        "No temporary file system is present.  Either call this method from within the "
                        "context of a task or surround with a 'with LocalTestFileSystem():' block.  Or "
                        "specify a path when calling this function.  Note: Cleanup is not automatic when a "
                        "path is specified."
                    )
                self._directory = _utils.AutoDeletingTempDir(
                    _uuid.uuid4().hex,
                    tmp_dir=_data_proxy.LocalWorkingDirectoryContext.get().name,
                )
                self._is_managed = True
                self._directory.__enter__()
                # TODO: Introduce system logging
                # logging.info("Copying recursively {} -> {}".format(self.remote_location, self.local_path))
                _data_proxy.Data.get_data(self.remote_location, self.local_path, is_multipart=True)

            # Read the files into blobs in case-insensitive lexicographically ascending orders
            self._blobs = []
            file_handles = []
            for local_path in sorted(self._directory.list_dir(), key=lambda x: x.lower()):
                b = Blob(
                    _os.path.join(self.remote_location, _os.path.basename(local_path)),
                    mode=self.mode,
                )
                b._local_path = local_path
                file_handles.append(b.__enter__())
                self._blobs.append(b)

            return file_handles
        except Exception:
            # Exit is idempotent so close partially opened context that way
            exc_type, exc_obj, exc_tb = _sys.exc_info()
            self.__exit__(exc_type, exc_obj, exc_tb)
            raise

    @_exception_scopes.system_entry_point
    def __exit__(self, exc_type, exc_val, exc_tb):
        for blob in self._blobs:
            blob.__exit__(exc_type, exc_val, exc_tb)
        self._blobs = []
        if self._is_managed:
            self._directory.__exit__(exc_type, exc_val, exc_tb)
            self._directory = None
            self._is_managed = False
        return False

    @property
    def local_path(self):
        """
        Local filesystem path where the file was downloaded
        :rtype: Text
        """
        if not self._directory:
            return None
        return self._directory.name

    @property
    def remote_location(self):
        """
        Path to where this MultiPartBlob will be synced.
        :rtype: Text
        """
        return self.uri

    @property
    def mode(self):
        """
        The mode string the MultiPartBlob is associated with.
        :rtype: Text
        """
        return self._mode

    @_exception_scopes.system_entry_point
    def create_part(self, name=None):
        """
        Method which will return a Blob object for writing into a multi-part blob.

        :param Text name: [optional] If name is provided, it is a specific partition name to place as part of the
            multipart blob.  When we read blobs from a multipart, it will be read in lexicographic order so this can be
            used to enforce ordering.  If not provided, the name is randomly generated.
        :rtype: Blob
        """
        if "w" not in self.mode:
            raise _user_exceptions.FlyteAssertion("Cannot create a blob in a read-only multipart blob")
        if name is None:
            name = _uuid.uuid4().hex
        if ":" in name or "/" in name:
            raise _user_exceptions.FlyteAssertion(
                name,
                "Cannot create a part of a multi-part object with ':' or '/' in the name.",
            )
        return Blob.create_at_known_location(
            _os.path.join(self.remote_location, name),
            mode=self.mode,
            format=self.metadata.type.format,
        )

    @_exception_scopes.system_entry_point
    def download(self, local_path=None, overwrite=False):
        """
        Forces the download of the remote multi-part blob to the local machine.
        :param Text local_path: [Optional] If provided, the blob pieces will be downloaded to this path.  This will
            make the resulting file objects unmanaged and it will not be cleaned up by the system upon exiting the
            context.
        :param bool overwrite: If true and local_path is specified, we will download the blob pieces and
            overwrite any existing files at that location.  Default is False.
        """
        if "r" not in self.mode:
            raise _user_exceptions.FlyteAssertion("Cannot download a write-only object!")

        if local_path:
            self._is_managed = False
        elif _data_proxy.LocalWorkingDirectoryContext.get() is None:
            raise _user_exceptions.FlyteAssertion(
                "No temporary file system is present.  Either call this method from within the "
                "context of a task or surround with a 'with LocalTestFileSystem():' block.  Or "
                "specify a path when calling this function.  Note: Cleanup is not automatic when a "
                "path is specified."
            )
        else:
            local_path = _data_proxy.LocalWorkingDirectoryContext.get().get_named_tempfile(_uuid.uuid4().hex)

        path_exists = _os.path.exists(local_path)
        if not path_exists or overwrite:
            if path_exists:
                _shutil.rmtree(local_path)
            _os.makedirs(local_path)
            self._directory = _utils.Directory(local_path)
            _data_proxy.Data.get_data(self.remote_location, self.local_path, is_multipart=True)
        else:
            raise _user_exceptions.FlyteAssertion(
                "Cannot download multi-part blob to a location that already exists when overwrite is not set to True. "
                "Attempted download from {} -> {}".format(self.remote_location, self.local_path)
            )

    @_exception_scopes.system_entry_point
    def upload(self):
        """
        Upload the multi-part blob to the remote location
        """
        if "w" not in self.mode:
            raise _user_exceptions.FlyteAssertion("Cannot upload a read-only multi-part blob!")

        elif not self.local_path:
            raise _user_exceptions.FlyteAssertion(
                "The multi-part blob is not currently backed by a local directoru "
                "and therefore cannot be uploaded.  Please write to this before "
                "attempting an upload."
            )
        else:
            # TODO: Introduce system logging
            # logging.info("Putting {} -> {}".format(self.local_path, self.remote_location))
            _data_proxy.Data.put_data(self.local_path, self.remote_location, is_multipart=True)



import sys

from setuptools import find_packages, setup  # noqa

# from flytekit.tools.lazy_loader import LazyLoadPlugin  # noqa
# extras_require = LazyLoadPlugin.get_extras_require()

MIN_PYTHON_VERSION = (3, 7)
CURRENT_PYTHON = sys.version_info[:2]
if CURRENT_PYTHON < MIN_PYTHON_VERSION:
    print("Minimum Python version is {}, but you are on {}".format(MIN_PYTHON_VERSION, CURRENT_PYTHON))
    sys.exit(-1)


spark = ["pyspark>=2.4.0,<3.0.0"]
spark3 = ["pyspark>=3.0.0"]
sidecar = ["k8s-proto>=0.0.3,<1.0.0"]
schema = ["numpy>=1.14.0,<2.0.0", "pandas>=0.22.0,<2.0.0", "pyarrow>2.0.0,<4.0.0"]
hive_sensor = ["hmsclient>=0.0.1,<1.0.0"]
notebook = ["papermill>=1.2.0", "nbconvert>=6.0.7", "ipykernel>=5.0.0"]
sagemaker = ["sagemaker-training>=3.6.2,<4.0.0"]

all_but_spark = sidecar + schema + hive_sensor + notebook + sagemaker

extras_require = {
    "spark": spark,
    "spark3": spark3,
    "sidecar": sidecar,
    "schema": schema,
    "hive_sensor": hive_sensor,
    "notebook": notebook,
    "sagemaker": sagemaker,
    "all-spark2.4": spark + all_but_spark,
    "all": spark3 + all_but_spark,
}

__version__ = "0.19.0b0"

setup(
    name="flytekit",
    version=__version__,
    maintainer="Flyte Org",
    maintainer_email="admin@flyte.org",
    packages=find_packages(exclude=["tests*"]),
    url="https://github.com/lyft/flytekit",
    description="Flyte SDK for Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "pyflyte-execute=flytekit.bin.entrypoint:execute_task_cmd",
            "pyflyte-fast-execute=flytekit.bin.entrypoint:fast_execute_task_cmd",
            "pyflyte-map-execute=flytekit.bin.entrypoint:map_execute_task_cmd",
            "pyflyte=flytekit.clis.sdk_in_container.pyflyte:main",
            "flyte-cli=flytekit.clis.flyte_cli.main:_flyte_cli",
        ]
    },
    install_requires=[
        "flyteidl>=0.18.26,<1.0.0",
        "wheel>=0.30.0,<1.0.0",
        "pandas>=1.0.0,<2.0.0",
        "pyarrow>=2.0.0,<4.0.0",
        "click>=6.6,<8.0",
        "croniter>=0.3.20,<4.0.0",
        "deprecated>=1.0,<2.0",
        "python-dateutil<=2.8.1,>=2.1",
        "grpcio>=1.3.0,<2.0",
        "protobuf>=3.6.1,<4",
        "pytimeparse>=1.1.8,<2.0.0",
        "pytz>=2017.2,<2018.5",
        "keyring>=18.0.1",
        "requests>=2.18.4,<3.0.0",
        "responses>=0.10.7",
        "six>=1.9.0,<2.0.0",
        "sortedcontainers>=1.5.9<3.0.0",
        "statsd>=3.0.0,<4.0.0",
        "urllib3>=1.22,<1.26",
        "wrapt>=1.0.0,<2.0.0",
        "retry==0.9.2",
        "dataclasses-json>=0.5.2",
        "natsort>=7.0.1",
        "dirhash>=0.2.1",
        "docker-image-py>=0.1.10",
    ],
    extras_require=extras_require,
    scripts=[
        "scripts/flytekit_build_image.sh",
        "scripts/flytekit_venv",
        "scripts/flytekit_sagemaker_runner.py",
        "flytekit/bin/entrypoint.py",
    ],
    license="apache2",
    python_requires=">=3.7",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from flytekit.configuration import common as _config_common

CLIENT_ID = _config_common.FlyteStringConfigurationEntry("credentials", "client_id", default=None)
"""
This is the public identifier for the app which handles authorization for a Flyte deployment.
More details here: https://www.oauth.com/oauth2-servers/client-registration/client-id-secret/.
"""

REDIRECT_URI = _config_common.FlyteStringConfigurationEntry(
    "credentials", "redirect_uri", default="http://localhost:12345/callback"
)
"""
This is the callback uri registered with the app which handles authorization for a Flyte deployment.
Please note the hardcoded port number. Ideally we would not do this, but some IDPs do not allow wildcards for
the URL, which means we have to use the same port every time. This is the only reason this is a configuration option,
otherwise, we'd just hardcode the callback path as a constant.
FYI, to see if a given port is already in use, run `sudo lsof -i :<port>` if on a Linux system.
More details here: https://www.oauth.com/oauth2-servers/redirect-uris/.
"""

AUTHORIZATION_METADATA_KEY = _config_common.FlyteStringConfigurationEntry(
    "credentials", "authorization_metadata_key", default="authorization"
)
"""
The authorization metadata key used for passing access tokens in gRPC requests.
Traditionally this value is 'authorization' however it is made configurable.
"""


CLIENT_CREDENTIALS_SECRET = _config_common.FlyteStringConfigurationEntry("credentials", "client_secret", default=None)
"""
Used for basic auth, which is automatically called during pyflyte. This will allow the Flyte engine to read the
password directly from the environment variable. Note that this is less secure! Please only use this if mounting the
secret as a file is impossible.
"""


CLIENT_CREDENTIALS_SCOPE = _config_common.FlyteStringConfigurationEntry("credentials", "scope", default=None)
"""
Used for basic auth, which is automatically called during pyflyte. This is the scope that will be requested. Because
there is no user explicitly in this auth flow, certain IDPs require a custom scope for basic auth in the configuration
of the authorization server.
"""

AUTH_MODE = _config_common.FlyteStringConfigurationEntry("credentials", "auth_mode", default="standard")
"""
The auth mode defines the behavior used to request and refresh credentials. The currently supported modes include:
- 'standard' This uses the pkce-enhanced authorization code flow by opening a browser window to initiate credentials
        access.
- 'basic' This uses cert-based auth in which the end user enters his/her username and password and public key encryption
        is used to facilitate authentication.
- None: No auth will be attempted.
"""



"""
Flytekit Directory Type (:mod:`flytekit.types.directory`)
==========================================================
.. currentmodule:: flytekit.types.directory

Similar to :py:class:`flytekit.types.file.FlyteFile` there are some 'preformatted' directory types.

.. autosummary::
   :toctree: generated/

   FlyteDirectory
   TensorboardLogs
"""

import typing

from .types import FlyteDirectory

# The following section provides some predefined aliases for commonly used FlyteDirectory formats.

TensorboardLogs = FlyteDirectory[typing.TypeVar("tensorboard")]
"""
    This type can be used to denote that the output is a folder that contains logs that can be loaded in tensorboard.
    this is usually the SummaryWriter output in pytorch or Keras callbacks which record the history readable by
    tensorboard
"""



from flyteidl.plugins import qubole_pb2 as _qubole

from flytekit.models import common as _common


class HiveQuery(_common.FlyteIdlEntity):
    def __init__(self, query, timeout_sec, retry_count):
        """
        Initializes a new HiveQuery.

        :param Text query: The query string.
        :param int timeout_sec:
        :param int retry_count:

        """
        self._query = query
        self._timeout_sec = timeout_sec
        self._retry_count = retry_count

    @property
    def query(self):
        """
        The query string.
        :rtype: str
        """
        return self._query

    @property
    def timeout_sec(self):
        """
        :rtype: int
        """
        return self._timeout_sec

    @property
    def retry_count(self):
        """
        :rtype: int
        """
        return self._retry_count

    def to_flyte_idl(self):
        """
        :rtype: _qubole.HiveQuery
        """
        return _qubole.HiveQuery(query=self.query, timeout_sec=self.timeout_sec, retryCount=self.retry_count)

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param _qubole.HiveQuery pb2_object:
        :return: HiveQuery
        """
        return cls(
            query=pb2_object.query,
            timeout_sec=pb2_object.timeout_sec,
            retry_count=pb2_object.retryCount,
        )


class HiveQueryCollection(_common.FlyteIdlEntity):
    def __init__(self, queries):
        """
        Initializes a new HiveQueryCollection.

        :param list[HiveQuery] queries: Queries to execute.
        """
        self._queries = queries

    @property
    def queries(self):
        """
        :rtype: list[HiveQuery]
        """
        return self._queries

    def to_flyte_idl(self):
        """
        :rtype: _qubole.HiveQueryCollection
        """
        return _qubole.HiveQueryCollection(
            queries=[query.to_flyte_idl() for query in self.queries] if self.queries else None
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param _qubole.HiveQuery pb2_object:
        :rtype: HiveQueryCollection
        """
        return cls(queries=[HiveQuery.from_flyte_idl(query) for query in pb2_object.queries])


class QuboleHiveJob(_common.FlyteIdlEntity):
    def __init__(self, query, cluster_label, tags, query_collection=None):
        """
        Initializes a HiveJob.

        :param HiveQuery query: Single query to execute
        :param Text cluster_label: The qubole cluster label to execute the query on
        :param list[Text] tags: User tags for the queries
        :param HiveQueryCollection query_collection: Deprecated Queries to execute.
        """
        self._query = query
        self._cluster_label = cluster_label
        self._tags = tags
        self._query_collection = query_collection

    @property
    def query_collection(self):
        """
        The queries to be executed
        :rtype: HiveQueryCollection
        """
        return self._query_collection

    @property
    def query(self):
        """
        The query to be executed
        :rtype: HiveQuery
        """
        return self._query

    @property
    def cluster_label(self):
        """
        The cluster label where the query should be executed
        :rtype: Text
        """
        return self._cluster_label

    @property
    def tags(self):
        """
        User tags for the queries
        :rtype: list[Text]
        """
        return self._tags

    def to_flyte_idl(self):
        """
        :rtype: _qubole.QuboleHiveJob
        """
        return _qubole.QuboleHiveJob(
            query_collection=self._query_collection.to_flyte_idl() if self._query_collection else None,
            query=self._query.to_flyte_idl() if self._query else None,
            cluster_label=self._cluster_label,
            tags=self._tags,
        )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param _qubole.QuboleHiveJob p:
        :rtype: QuboleHiveJob
        """
        return cls(
            query_collection=HiveQueryCollection.from_flyte_idl(p.query_collection)
            if p.HasField("query_collection")
            else None,
            query=HiveQuery.from_flyte_idl(p.query) if p.HasField("query") else None,
            cluster_label=p.cluster_label,
            tags=p.tags,
        )



from abc import ABC

from flytekit.core.tracker import TrackedInstance


class FlyteTrackedABC(type(TrackedInstance), type(ABC)):
    """
    This class exists because if you try to inherit from abc.ABC and TrackedInstance by itself, you'll get the
    well-known ``TypeError: metaclass conflict: the metaclass of a derived class must be a (non-strict) subclass
    of the metaclasses of all its bases`` error.
    """



from contextlib import contextmanager
from typing import Union
from unittest.mock import MagicMock

from flytekit.core.base_task import PythonTask
from flytekit.core.reference_entity import ReferenceEntity
from flytekit.core.workflow import WorkflowBase
from flytekit.loggers import logger


@contextmanager
def task_mock(t: PythonTask) -> MagicMock:
    """
    Use this method to mock a task declaration. It can mock any Task in Flytekit as long as it has a python native
    interface associated with it.

    The returned object is a MagicMock and allows to perform all such methods. This MagicMock, mocks the execute method
    on the PythonTask

    Usage:

        .. code-block:: python

            @task
            def t1(i: int) -> int:
               pass

            with task_mock(t1) as m:
               m.side_effect = lambda x: x
               t1(10)
               # The mock is valid only within this context
    """

    if not isinstance(t, PythonTask) and not isinstance(t, WorkflowBase) and not isinstance(t, ReferenceEntity):
        raise Exception("Can only be used for tasks")

    m = MagicMock()

    def _log(*args, **kwargs):
        logger.warning(f"Invoking mock method for task: '{t.name}'")
        return m(*args, **kwargs)

    _captured_fn = t.execute
    t.execute = _log
    yield m
    t.execute = _captured_fn


def patch(target: Union[PythonTask, WorkflowBase, ReferenceEntity]):
    """
    This is a decorator used for testing.
    """
    if (
        not isinstance(target, PythonTask)
        and not isinstance(target, WorkflowBase)
        and not isinstance(target, ReferenceEntity)
    ):
        raise Exception("Can only use mocks on tasks/workflows declared in Python.")

    logger.info(
        "When using this patch function on Flyte entities, please be aware weird issues may arise if also"
        "using mock.patch on internal Flyte classes like PythonFunctionWorkflow. See"
        "https://github.com/flyteorg/flyte/issues/854 for more information"
    )

    def wrapper(test_fn):
        def new_test(*args, **kwargs):
            logger.warning(f"Invoking mock method for target: '{target.name}'")
            m = MagicMock()
            saved = target.execute
            target.execute = m
            results = test_fn(m, *args, **kwargs)
            target.execute = saved
            return results

        return new_test

    return wrapper



import os as _os
from typing import List as _List

import click

from flytekit.clis.sdk_in_container.constants import CTX_DOMAIN, CTX_PACKAGES, CTX_PROJECT, CTX_TEST
from flytekit.common import utils as _utils
from flytekit.common.core import identifier as _identifier
from flytekit.common.tasks import sdk_runnable as _sdk_runnable_task
from flytekit.common.tasks import task as _task
from flytekit.configuration import sdk as _sdk_config
from flytekit.tools.fast_registration import compute_digest as _compute_digest
from flytekit.tools.fast_registration import get_additional_distribution_loc as _get_additional_distribution_loc
from flytekit.tools.fast_registration import upload_package as _upload_package
from flytekit.tools.module_loader import iterate_registerable_entities_in_order


def fast_register_all(
    project: str,
    domain: str,
    pkgs: _List[str],
    test: bool,
    version: str,
    source_dir: _os.PathLike,
    dest_dir: _os.PathLike = None,
):
    if test:
        click.echo("Test switch enabled, not doing anything...")

    if not version:
        digest = _compute_digest(source_dir)
    else:
        digest = version
    remote_package_path = _upload_package(source_dir, digest, _sdk_config.FAST_REGISTRATION_DIR.get())

    click.echo(
        "Running task, workflow, and launch plan fast registration for {}, {}, {} with version {} and code dir {}".format(
            project, domain, pkgs, digest, source_dir
        )
    )

    # m = module (i.e. python file)
    # k = value of dir(m), type str
    # o = object (e.g. SdkWorkflow)
    for m, k, o in iterate_registerable_entities_in_order(pkgs):
        name = _utils.fqdn(m.__name__, k, entity_type=o.resource_type)
        o._id = _identifier.Identifier(o.resource_type, project, domain, name, digest)

        if test:
            click.echo("Would fast register {:20} {}".format("{}:".format(o.entity_type_text), o.id.name))
        else:
            click.echo("Fast registering {:20} {}".format("{}:".format(o.entity_type_text), o.id.name))
            _get_additional_distribution_loc(_sdk_config.FAST_REGISTRATION_DIR.get(), digest)
            if isinstance(o, _sdk_runnable_task.SdkRunnableTask):
                o.fast_register(project, domain, o.id.name, digest, remote_package_path, dest_dir)
            else:
                o.register(project, domain, o.id.name, digest)


def fast_register_tasks_only(
    project: str,
    domain: str,
    pkgs: _List[str],
    test: bool,
    version: str,
    source_dir: _os.PathLike,
    dest_dir: _os.PathLike = None,
):
    if test:
        click.echo("Test switch enabled, not doing anything...")

    if not version:
        digest = _compute_digest(source_dir)
    else:
        digest = version
    remote_package_path = _upload_package(source_dir, digest, _sdk_config.FAST_REGISTRATION_DIR.get())

    click.echo(
        "Running task only fast registration for {}, {}, {} with version {} and code dir {}".format(
            project, domain, pkgs, digest, source_dir
        )
    )

    # Discover all tasks by loading the module
    for m, k, t in iterate_registerable_entities_in_order(pkgs, include_entities={_task.SdkTask}):
        name = _utils.fqdn(m.__name__, k, entity_type=t.resource_type)

        if test:
            click.echo("Would fast register task {:20} {}".format("{}:".format(t.entity_type_text), name))
        else:
            click.echo("Fast registering task {:20} {}".format("{}:".format(t.entity_type_text), name))
            if isinstance(t, _sdk_runnable_task.SdkRunnableTask):
                t.fast_register(project, domain, name, digest, remote_package_path, dest_dir)
            else:
                t.register(project, domain, name, digest)


@click.group("fast-register")
@click.option("--test", is_flag=True, help="Dry run, do not actually register with Admin")
@click.pass_context
def fast_register(ctx, test=None):
    """
    Run fast registration steps for the Flyte entities in this container. This is an optimization to avoid the
    conventional container build and upload cycle. This can be useful for fast iteration when making code changes.
    If you do need to change the container itself (e.g. by adding a new dependency/import) you must rebuild and
    upload a container.

    Caveats: Your flyte config must specify a fast registration dir like so:
    [sdk]
    fast_registration_dir=s3://my-s3-bucket/dir

    **and** ensure that the role specified in [auth] section of your config has read access to this remote location.
    Furthermore, the role you assume to call fast-register must have **write** permission to this remote location.

    Run with the --test switch for a dry run to see what will be registered.  A default launch plan will also be
    created, if a role can be found in the environment variables.
    """

    ctx.obj[CTX_TEST] = test


@click.command("tasks")
@click.option(
    "--source-dir",
    type=str,
    help="The root dir of the code that should be uploaded for fast registration.",
    required=True,
)
@click.option(
    "--dest-dir",
    type=str,
    help="[Optional] The output directory of code which is downloaded during fast registration. "
    "If the current working directory at the time of installation is not desired",
)
@click.option(
    "-v",
    "--version",
    type=str,
    help="Version to register tasks with. This is normally computed deterministically from your code, "
    "but you can override here.",
)
@click.pass_context
def tasks(ctx, source_dir, dest_dir=None, version=None):
    """
    Only fast register tasks.

    For example, consider a sample directory where tasks defined in workflows/ imports code from util/ like so:

    \b
    $ tree /root/code/
    /root/code/
    ├── Dockerfile
    ├── Makefile
    ├── README.md
    ├── conf.py
    ├── notebook.config
    ├── workflows
    │   ├── __init__.py
    │   ├── compose
    │   │   ├── README.md
    │   │   ├── __init__.py
    │   │   ├── a_workflow.py
    │   │   ├── b_workflow.py
    ├── util
    │   ├── __init__.py
    │   ├── shared_task_code.py
    ├── requirements.txt
    ├── flyte.config

    Your source dir will need to be /root/code/ rather than the workflow packages dir /root/code/workflows you might
    have specified in your flyte.config because all of the code your workflows depends on needs to be encapsulated in
    `source_dir`, like so:

    pyflyte -p myproject -d development fast-register tasks --source-dir /root/code/

    """
    project = ctx.obj[CTX_PROJECT]
    domain = ctx.obj[CTX_DOMAIN]
    test = ctx.obj[CTX_TEST]
    pkgs = ctx.obj[CTX_PACKAGES]

    fast_register_tasks_only(project, domain, pkgs, test, version, source_dir, dest_dir)


@click.command("workflows")
@click.option(
    "--source-dir",
    type=str,
    help="The root dir of the code that should be uploaded for fast registration.",
    required=True,
)
@click.option(
    "--dest-dir",
    type=str,
    help="[Optional] The output directory of code which is downloaded during fast registration. "
    "If the current working directory at the time of installation is not desired",
)
@click.option(
    "-v",
    "--version",
    type=str,
    help="Version to register entities with. This is normally computed deterministically from your code, "
    "but you can override here.",
)
@click.pass_context
def workflows(ctx, source_dir, dest_dir=None, version=None):
    """
    Fast register both tasks and workflows.  Also create and register a default launch plan for all workflows.
    The `source_dir` param should point to the root directory of your project that contains all of your working code.

    For example, consider a sample directory structure where code in workflows/ imports code from util/ like so:

    \b
    $ tree /root/code/
    /root/code/
    ├── Dockerfile
    ├── Makefile
    ├── README.md
    ├── conf.py
    ├── notebook.config
    ├── workflows
    │   ├── __init__.py
    │   ├── compose
    │   │   ├── README.md
    │   │   ├── __init__.py
    │   │   ├── a_workflow.py
    │   │   ├── b_workflow.py
    ├── util
    │   ├── __init__.py
    │   ├── shared_workflow_code.py
    ├── requirements.txt
    ├── flyte.config

    Your source dir will need to be /root/code/ rather than the workflow packages dir /root/code/workflows you might
    have specified in your flyte.config because all of the code your workflows depends on needs to be encapsulated in
    `source_dir`, like so:

    pyflyte -p myproject -d development fast-register workflows --source-dir /root/code/
    """
    project = ctx.obj[CTX_PROJECT]
    domain = ctx.obj[CTX_DOMAIN]
    test = ctx.obj[CTX_TEST]
    pkgs = ctx.obj[CTX_PACKAGES]

    fast_register_all(project, domain, pkgs, test, version, source_dir, dest_dir)


fast_register.add_command(tasks)
fast_register.add_command(workflows)



# This package has been developed by the following github.com users and adopted to this repo.
# @yashlyft, @tildedave, @jmphili, @pgerstoft, @johnliu, @ikonst, @mxr, @simonj2, @rowilia, @asottile, @mattdcamp



import os
import tempfile

import pytest
from flytekitplugins.awssagemaker.distributed_training import setup_envars_for_testing
from flytekitplugins.awssagemaker.training import SagemakerBuiltinAlgorithmsTask, SagemakerTrainingJobConfig

import flytekit
from flytekit import task
from flytekit.common.tasks.sdk_runnable import ExecutionParameters
from flytekit.extend import Image, ImageConfig, SerializationSettings
from flytekit.models.sagemaker.training_job import (
    AlgorithmName,
    AlgorithmSpecification,
    DistributedProtocol,
    TrainingJobResourceConfig,
)


def _get_reg_settings():
    default_img = Image(name="default", fqn="test", tag="tag")
    settings = SerializationSettings(
        project="project",
        domain="domain",
        version="version",
        env={"FOO": "baz"},
        image_config=ImageConfig(default_image=default_img, images=[default_img]),
    )
    return settings


def test_builtin_training():
    trainer = SagemakerBuiltinAlgorithmsTask(
        name="builtin-trainer",
        task_config=SagemakerTrainingJobConfig(
            training_job_resource_config=TrainingJobResourceConfig(
                instance_count=1,
                instance_type="ml-xlarge",
                volume_size_in_gb=1,
            ),
            algorithm_specification=AlgorithmSpecification(
                algorithm_name=AlgorithmName.XGBOOST,
            ),
        ),
    )

    assert trainer.python_interface.inputs.keys() == {"static_hyperparameters", "train", "validation"}
    assert trainer.python_interface.outputs.keys() == {"model"}

    with tempfile.TemporaryDirectory() as tmp:
        x = os.path.join(tmp, "x")
        y = os.path.join(tmp, "y")
        with open(x, "w") as f:
            f.write("test")
        with open(y, "w") as f:
            f.write("test")
        with pytest.raises(NotImplementedError):
            trainer(static_hyperparameters={}, train=x, validation=y)

    assert trainer.get_custom(_get_reg_settings()) == {
        "algorithmSpecification": {"algorithmName": "XGBOOST"},
        "trainingJobResourceConfig": {"instanceCount": "1", "instanceType": "ml-xlarge", "volumeSizeInGb": "1"},
    }


def test_custom_training():
    @task(
        task_config=SagemakerTrainingJobConfig(
            training_job_resource_config=TrainingJobResourceConfig(
                instance_type="ml-xlarge",
                volume_size_in_gb=1,
            ),
            algorithm_specification=AlgorithmSpecification(
                algorithm_name=AlgorithmName.CUSTOM,
            ),
        )
    )
    def my_custom_trainer(x: int) -> int:
        return x

    assert my_custom_trainer.python_interface.inputs == {"x": int}
    assert my_custom_trainer.python_interface.outputs == {"o0": int}

    assert my_custom_trainer(x=10) == 10

    assert my_custom_trainer.get_custom(_get_reg_settings()) == {
        "algorithmSpecification": {},
        "trainingJobResourceConfig": {"instanceCount": "1", "instanceType": "ml-xlarge", "volumeSizeInGb": "1"},
    }


def test_distributed_custom_training():
    setup_envars_for_testing()

    @task(
        task_config=SagemakerTrainingJobConfig(
            training_job_resource_config=TrainingJobResourceConfig(
                instance_type="ml-xlarge",
                volume_size_in_gb=1,
                instance_count=2,  # Indicates distributed training
                distributed_protocol=DistributedProtocol.MPI,
            ),
            algorithm_specification=AlgorithmSpecification(
                algorithm_name=AlgorithmName.CUSTOM,
            ),
        )
    )
    def my_custom_trainer(x: int) -> int:
        assert flytekit.current_context().distributed_training_context is not None
        return x

    assert my_custom_trainer.python_interface.inputs == {"x": int}
    assert my_custom_trainer.python_interface.outputs == {"o0": int}

    assert my_custom_trainer(x=10) == 10

    assert my_custom_trainer._is_distributed() is True

    pb = ExecutionParameters.new_builder()
    pb.working_dir = "/tmp"
    p = pb.build()
    new_p = my_custom_trainer.pre_execute(p)
    assert new_p is not None
    assert new_p.has_attr("distributed_training_context")

    assert my_custom_trainer.get_custom(_get_reg_settings()) == {
        "algorithmSpecification": {},
        "trainingJobResourceConfig": {
            "distributedProtocol": "MPI",
            "instanceCount": "2",
            "instanceType": "ml-xlarge",
            "volumeSizeInGb": "1",
        },
    }



"""
=====================
Core Flytekit
=====================

.. currentmodule:: flytekit

This package contains all of the most common abstractions you'll need to write Flyte workflows, and extend Flytekit.

Basic Authoring
===============

These are the essentials needed to get started writing tasks and workflows. The elements here correspond well with :std:ref:`Basic <cookbook:sphx_glr_auto_core_flyte_basics>` section of the cookbook.

.. autosummary::
   :nosignatures:
   :toctree: generated/

   task - This is the basic decorator to use to turn a properly type-annotated function into a Flyte task.
   workflow - This will turn a properly type-annotated function into a Flyte workflow.
   kwtypes - Helper function that makes it slightly easier to declare certain functions.
   current_context - This function returns an ExecutionParameters object.
   ExecutionParameters - This object gives the user at Task run time useful information about the execution environment.
   FlyteContext - This is an flytekit-internal context that is sometimes useful for end-users to have access to.


Running Locally
------------------
Tasks and Workflows can both be locally run (assuming the relevant tasks are capable of local execution). This is useful for unit testing.


Branching and Conditionals
==========================

Branches and conditionals can be expressed explicitly in Flyte. These conditions are evaluated
in the flyte engine and hence should be used for control flow. ``dynamic workflows`` can be used to perform custom conditional logic not supported by flytekit

.. autosummary::
   :nosignatures:
   :toctree: generated/

   conditional


Customizing Tasks & Workflows
==============================

.. autosummary::
   :nosignatures:
   :toctree: generated/

   TaskMetadata - Wrapper object that allows users to specify Task
   Resources - Things like CPUs/Memory, etc.
   WorkflowFailurePolicy - Customizes what happens when a workflow fails.


Dynamic and Nested Workflows
==============================
See the :py:mod:`Dynamic <flytekit.core.dynamic_workflow_task>` module for more information.

.. autosummary::
   :nosignatures:
   :toctree: generated/

   dynamic

Scheduling and Notifications
============================

See the :py:mod:`Notifications Module <flytekit.core.notification>` and
:py:mod:`Schedules Module <flytekit.core.schedule>` for more information.

.. autosummary::
   :nosignatures:
   :toctree: generated/

   CronSchedule
   FixedRate
   Email
   PagerDuty
   Slack

Reference Entities
====================

.. autosummary::
   :nosignatures:
   :toctree: generated/

   get_reference_entity
   LaunchPlanReference
   TaskReference
   WorkflowReference
   reference_task
   reference_workflow

Core Task Types
=================

.. autosummary::
   :nosignatures:
   :toctree: generated/

   SQLTask
   ContainerTask
   PythonFunctionTask
   PythonInstanceTask
   LaunchPlan

Secrets and SecurityContext
============================

.. autosummary::
   :nosignatures:
   :toctree: generated/

   Secret
   SecurityContext

Core Modules
=============

.. autosummary::
   :nosignatures:
   :toctree: generated/

   core.dynamic_workflow_task
   core.notification
   core.schedule

"""


import flytekit.plugins  # This will be deprecated, these are the old plugins, the new plugins live in plugins/
from flytekit.core.base_sql_task import SQLTask
from flytekit.core.base_task import SecurityContext, TaskMetadata, kwtypes
from flytekit.core.condition import conditional
from flytekit.core.container_task import ContainerTask
from flytekit.core.context_manager import ExecutionParameters, FlyteContext
from flytekit.core.dynamic_workflow_task import dynamic
from flytekit.core.launch_plan import LaunchPlan
from flytekit.core.map_task import map_task
from flytekit.core.notification import Email, PagerDuty, Slack
from flytekit.core.python_function_task import PythonFunctionTask, PythonInstanceTask
from flytekit.core.reference import get_reference_entity
from flytekit.core.reference_entity import LaunchPlanReference, TaskReference, WorkflowReference
from flytekit.core.resources import Resources
from flytekit.core.schedule import CronSchedule, FixedRate
from flytekit.core.task import Secret, reference_task, task
from flytekit.core.workflow import ImperativeWorkflow as Workflow
from flytekit.core.workflow import WorkflowFailurePolicy, reference_workflow, workflow
from flytekit.loggers import logger
from flytekit.types import schema

__version__ = "0.19.0b0"


def current_context() -> ExecutionParameters:
    """
    Use this method to get a handle of specific parameters available in a flyte task.

    Usage

    .. code-block::

        flytekit.current_context().logging.info(...)

    Available params are documented in :py:class:`flytekit.core.context_manager.ExecutionParams`.
    There are some special params, that should be available
    """
    return FlyteContext.current_context().user_space_params



import datetime as _datetime
import logging


class MockStats(object):
    def __init__(self, scope="", tags=None):
        """
        Initializes a new mock stats object
        :param Text scope:
        :param dict[Text, Text] tags:
        """
        self.scope = scope
        self.tags = tags
        self._records = {}
        self._records_tags = {}

    def incr(self, metric, count=1, tags=None, **kwargs):
        full_name = self.scope + "." + metric
        self._records[full_name] = self._records.get(full_name, 0) + count
        self._records_tags[full_name] = tags or {}

    def decr(self, metric, count=1, tags=None, **kwargs):
        full_name = self.scope + "." + metric
        self._records[full_name] = self._records.get(full_name, 0) - count
        self._records_tags[full_name] = tags or {}

    def timing(self, metric):
        logging.warning("mock timing isn't implemented yet.")

    def timer(self, metric, tags=None, **kwargs):
        return _Timer(self, metric, tags=tags or {})

    def gauge(self, metric, value, tags=None, **kwargs):
        full_name = self.scope + "." + metric
        self._records[full_name] = value
        self._records_tags[full_name] = tags or {}

    def current_value(self, metric):
        full_name = self.scope + "." + metric
        return self._records.get(full_name, None)

    def current_tags(self, metric):
        full_name = self.scope + "." + metric
        return self._records_tags.get(full_name, None)


class _Timer(object):
    def __init__(self, mock_stats, metric, tags):
        """
        :param MockStats mock_stats:
        :param Text metric:
        """
        self._mock_stats = mock_stats
        self._metric = metric
        self._timer = None
        self._tags = tags

    def __enter__(self):
        self._timer = _datetime.datetime.utcnow()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._mock_stats.gauge(self._metric, _datetime.datetime.utcnow() - self._timer, tags=self._tags)
        self._timer = None



from __future__ import annotations

import os
from typing import Any, Dict, List, Optional, Type, TypeVar

from flyteidl.core import tasks_pb2 as _tasks_pb2

from flytekit.common import utils as common_utils
from flytekit.common.tasks.raw_container import _get_container_definition
from flytekit.core.base_task import PythonTask, Task, TaskResolverMixin
from flytekit.core.context_manager import FlyteContext, Image, ImageConfig, SerializationSettings
from flytekit.core.resources import Resources, ResourceSpec
from flytekit.core.shim_task import ExecutableTemplateShimTask, ShimTaskExecutor
from flytekit.core.tracker import TrackedInstance
from flytekit.loggers import logger
from flytekit.models import task as _task_model
from flytekit.models.core import identifier as identifier_models
from flytekit.models.security import Secret, SecurityContext
from flytekit.tools.module_loader import load_object_from_module

TC = TypeVar("TC")


class PythonCustomizedContainerTask(ExecutableTemplateShimTask, PythonTask[TC]):
    """
    Please take a look at the comments for ``ExecutableTemplateShimTask`` as well. This class should be subclassed
    and a custom Executor provided as a default to this parent class constructor when building a new external-container
    flytekit-only plugin.

    This class provides authors of new task types the basic scaffolding to create task-template based tasks. In order
    to write such a task, authors need to:

      * subclass the ``ShimTaskExecutor`` class  and override the ``execute_from_model`` function. This function is
        where all the business logic should go. Keep in mind though that you, the plugin author, will not have access
        to anything that's not serialized within the ``TaskTemplate`` which is why you'll also need to
      * subclass this class, and override the ``get_custom`` function to include all the information the executor
        will need to run.
      * Also pass the executor you created as the ``executor_type`` argument of this class's constructor.

    Keep in mind that the total size of the ``TaskTemplate`` still needs to be small, since these will be accessed
    frequently by the Flyte engine.
    """

    SERIALIZE_SETTINGS = SerializationSettings(
        project="PLACEHOLDER_PROJECT",
        domain="LOCAL",
        version="PLACEHOLDER_VERSION",
        env=None,
        image_config=ImageConfig(
            default_image=Image(name="custom_container_task", fqn="flyteorg.io/placeholder", tag="image")
        ),
    )

    def __init__(
        self,
        name: str,
        task_config: TC,
        container_image: str,
        executor_type: Type[ShimTaskExecutor],
        task_resolver: Optional[TaskTemplateResolver] = None,
        task_type="python-task",
        requests: Optional[Resources] = None,
        limits: Optional[Resources] = None,
        environment: Optional[Dict[str, str]] = None,
        secret_requests: Optional[List[Secret]] = None,
        **kwargs,
    ):
        """
        :param name: unique name for the task, usually the function's module and name.
        :param task_config: Configuration object for Task. Should be a unique type for that specific Task
        :param container_image: This is the external container image the task should run at platform-run-time.
        :param executor: This is an executor which will actually provide the business logic.
        :param task_resolver: Custom resolver - if you don't make one, use the default task template resolver.
        :param task_type: String task type to be associated with this Task
        :param requests: custom resource request settings.
        :param limits: custom resource limit settings.
        :param environment: Environment variables you want the task to have when run.
        :param List[Secret] secret_requests: Secrets that are requested by this container execution. These secrets will
                                           be mounted based on the configuration in the Secret and available through
                                           the SecretManager using the name of the secret as the group
                                           Ideally the secret keys should also be semi-descriptive.
                                           The key values will be available from runtime, if the backend is configured
                       to provide secrets and if secrets are available in the configured secrets store.
                       Possible options for secret stores are
                        - `Vault <https://www.vaultproject.io/>`,
                        - `Confidant <https://lyft.github.io/confidant/>`,
                        - `Kube secrets <https://kubernetes.io/docs/concepts/configuration/secret/>`
                        - `AWS Parameter store <https://docs.aws.amazon.com/systems-manager/latest/userguide/systems-manager-parameter-store.html>`_
                        etc
        """
        sec_ctx = None
        if secret_requests:
            for s in secret_requests:
                if not isinstance(s, Secret):
                    raise AssertionError(f"Secret {s} should be of type flytekit.Secret, received {type(s)}")
            sec_ctx = SecurityContext(secrets=secret_requests)
        super().__init__(
            tt=None,
            executor_type=executor_type,
            task_type=task_type,
            name=name,
            task_config=task_config,
            security_ctx=sec_ctx,
            **kwargs,
        )
        self._resources = ResourceSpec(
            requests=requests if requests else Resources(), limits=limits if limits else Resources()
        )
        self._environment = environment
        self._container_image = container_image
        self._task_resolver = task_resolver or default_task_template_resolver

    def get_custom(self, settings: SerializationSettings) -> Dict[str, Any]:
        # Overriding base implementation to raise an error, force task author to implement
        raise NotImplementedError

    def get_config(self, settings: SerializationSettings) -> Dict[str, str]:
        # Overriding base implementation but not doing anything. Technically this should be the task config,
        # but the IDL limitation that the value also has to be a string is very limiting.
        # Recommend putting information you need in the config into custom instead, because when serializing
        # the custom field, we jsonify custom and the place it into a protobuf struct. This config field
        # just gets put into a Dict[str, str]
        return {}

    @property
    def resources(self) -> ResourceSpec:
        return self._resources

    @property
    def task_resolver(self) -> TaskTemplateResolver:
        return self._task_resolver

    @property
    def task_template(self) -> Optional[_task_model.TaskTemplate]:
        """
        Override the base class implementation to serialize on first call.
        """
        return self._task_template or self.serialize_to_model(settings=PythonCustomizedContainerTask.SERIALIZE_SETTINGS)

    @property
    def container_image(self) -> str:
        return self._container_image

    def get_command(self, settings: SerializationSettings) -> List[str]:
        container_args = [
            "pyflyte-execute",
            "--inputs",
            "{{.input}}",
            "--output-prefix",
            "{{.outputPrefix}}",
            "--raw-output-data-prefix",
            "{{.rawOutputDataPrefix}}",
            "--resolver",
            self.task_resolver.location,
            "--",
            *self.task_resolver.loader_args(settings, self),
        ]

        return container_args

    def get_container(self, settings: SerializationSettings) -> _task_model.Container:
        env = {**settings.env, **self.environment} if self.environment else settings.env
        return _get_container_definition(
            image=self.container_image,
            command=[],
            args=self.get_command(settings=settings),
            data_loading_config=None,
            environment=env,
            storage_request=self.resources.requests.storage,
            cpu_request=self.resources.requests.cpu,
            gpu_request=self.resources.requests.gpu,
            memory_request=self.resources.requests.mem,
            storage_limit=self.resources.limits.storage,
            cpu_limit=self.resources.limits.cpu,
            gpu_limit=self.resources.limits.gpu,
            memory_limit=self.resources.limits.mem,
        )

    def serialize_to_model(self, settings: SerializationSettings) -> _task_model.TaskTemplate:
        # This doesn't get called from translator unfortunately. Will need to move the translator to use the model
        # objects directly first.
        # Note: This doesn't settle the issue of duplicate registrations. We'll need to figure that out somehow.
        # TODO: After new control plane classes are in, promote the template to a FlyteTask, so that authors of
        #  customized-container tasks have a familiar thing to work with.
        obj = _task_model.TaskTemplate(
            identifier_models.Identifier(
                identifier_models.ResourceType.TASK, settings.project, settings.domain, self.name, settings.version
            ),
            self.task_type,
            self.metadata.to_taskmetadata_model(),
            self.interface,
            self.get_custom(settings),
            container=self.get_container(settings),
            config=self.get_config(settings),
        )
        self._task_template = obj
        return obj


class TaskTemplateResolver(TrackedInstance, TaskResolverMixin):
    """
    This is a special resolver that resolves the task above at execution time, using only the ``TaskTemplate``,
    meaning it should only be used for tasks that contain all pertinent information within the template itself.

    This class differs from some TaskResolverMixin pattern a bit. Most of the other resolvers you'll find,

    * restores the same task when ``load_task`` is called as the object that ``loader_args`` was called on.
      That is, even though at run time it's in a container on a cluster and is obviously a different Python process,
      the Python object in memory should look the same.
    * offers a one-to-one mapping between the list of strings returned by the ``loader_args`` function, an the task,
      at least within the container.

    This resolver differs in that,
    * when loading a task, the task that is a loaded is always an ``ExecutableTemplateShimTask``, regardless of what
      kind of task it was originally. It will only ever have what's available to it from the ``TaskTemplate``. No
      information that wasn't serialized into the template will be available.
    * all tasks will result in the same list of strings for a given subclass of the ``ShimTaskExecutor``
      executor. The strings will be ``["{{.taskTemplatePath}}", "path.to.your.executor"]``

    Also, ``get_all_tasks`` will always return an empty list, at least for now.
    """

    def __init__(self):
        super(TaskTemplateResolver, self).__init__()

    def name(self) -> str:
        return "task template resolver"

    # The return type of this function is different, it should be a Task, but it's not because it doesn't make
    # sense for ExecutableTemplateShimTask to inherit from Task.
    def load_task(self, loader_args: List[str]) -> ExecutableTemplateShimTask:
        logger.info(f"Task template loader args: {loader_args}")
        ctx = FlyteContext.current_context()
        task_template_local_path = os.path.join(ctx.execution_state.working_dir, "task_template.pb")
        ctx.file_access.get_data(loader_args[0], task_template_local_path)
        task_template_proto = common_utils.load_proto_from_file(_tasks_pb2.TaskTemplate, task_template_local_path)
        task_template_model = _task_model.TaskTemplate.from_flyte_idl(task_template_proto)

        executor_class = load_object_from_module(loader_args[1])
        return ExecutableTemplateShimTask(task_template_model, executor_class)

    def loader_args(self, settings: SerializationSettings, t: PythonCustomizedContainerTask) -> List[str]:
        return ["{{.taskTemplatePath}}", f"{t.executor_type.__module__}.{t.executor_type.__name__}"]

    def get_all_tasks(self) -> List[Task]:
        return []


default_task_template_resolver = TaskTemplateResolver()
